# textfx.py -- funky textual effects
# Copyright 2004 Joe Wreschnig <piman@sacredchao.net>
# Released under the terms of the GNU GPL v2.
__revision__ = "$Id: textfx.py 286 2004-09-04 03:51:59Z piman $"

import pygame
import math
import load

from constants import *

# From the Pygame font effect contest, by Pete Shinners. Displays
# some waving text.
class TextWavey(object):
    def __init__(self, size, message, fontcolor, amount = 10):
        self.base = shadow(message, size, color = fontcolor)
        self.steps = range(0, self.base.get_width(), 2)
        self.amount = amount
        self.size = self.base.get_rect().inflate(0, amount).size
        self.offset = 0.0

    def animate(self):
        s = pygame.Surface(self.size, SRCALPHA, 32)
        s.fill([0, 0, 0, 0])
        height = self.size[1]
        self.offset += 0.5
        for step in self.steps:
            src = Rect(step, 0, 2, height)
            dst = src.move(0, math.cos(self.offset + step*.02) * self.amount)
            s.blit(self.base, dst, src)
        return s

# A font capable of wrapping text at a given width.
class WrapFont(object):

    # size is the (integer) size of the font to create. width is the
    # maximum width of the rendered surface.
    def __init__(self, size, width):
        self._font = pygame.font.Font(None, size)
        self._width = width
        self._ls = self._font.get_linesize()
        self._ds = - self._font.get_descent()

    def get_linesize(self): return self._ls

    # Return the number of lines a bit of text will require.
    def lines(self, text):
        lines = 1
        words = text.split()
        start = 0
        for i in range(len(words)):
            line = " ".join(words[start:i+1])
            if self._font.size(line)[0] > self._width:
                lines += 1
                start = i
        return lines

    # Return the total size of Surface that render will return for
    # the given text.
    def size(self, text):
        return [self._width, self.lines(text) * self._ls + self._ds]
 
    # Render some text, optionally with shadowing.
    def render(self, text, color = [255, 255, 255], shdw = True):
        lines = []
        words = text.split()
        start = 0
        for i in range(len(words)):
            line = " ".join(words[start:i+1])
            if self._font.size(line)[0] > self._width:
                line = " ".join(words[start:i])
                if shdw: t = shadow(line, self._font, color)
                else: t = self._font.render(line, True, color)
                lines.append(t)
                start = i
        line = " ".join(words[start:])
        if shdw: t = shadow(line, self._font, color)
        else: t = self._font.render(line, True, color)
        lines.append(t)
 
        image = pygame.Surface([self._width,
                                len(lines) * self._ls + self._ds],
                               SRCALPHA, 32)
        image.fill([0, 0, 0, 0])
        for i, line in enumerate(lines):
            image.blit(line, [0, i * self._ls])
        return image

# Shadow some text (by 1 pixel). Resulting shadow is down and to
# the left, and 1/10th as bright.
def shadow(string, font, color = [255, 255, 255]):
    if isinstance(font, int): font = pygame.font.Font(None, font)
    t1 = font.render(string, True, color)
    t2 = font.render(string, True, [c // 10 for c in color])
    srf = pygame.Surface([t1.get_width() + 1, t1.get_height() + 1],
                         SRCALPHA, 32)
    srf.blit(t2, [1, 1])
    srf.blit(t1, [0, 0])
    return srf

# Render a box, and put a (shadowed) character inside it.
def lettered_box(char, clr):
    img = load.block(clr)
    t = shadow(char, 32)
    img.blit(t, t.get_rect(center = [15, 15]))
    return img

# Generate a pile of boxes from a string, split on spaces, one word
# per line. The color is random.
def boxes_from_string(string, pos = None):
    lines = string.split(" ")
    img = pygame.Surface([32 * max(map(len, lines)), 32 * len(lines)])
    colors = ["blue", "red", "green", "orange", "purple"]
    if pos is not None: clr = colors[pos % len(colors)]
    else: clr = random.choice(colors)

    for i, string in enumerate(lines):
        limg = pygame.Surface([32 * len(string), 32])
        for j, letter in enumerate(string):
            nimg = lettered_box(letter, clr)
            limg.blit(nimg, [j * 32, 0])
        r = limg.get_rect(midtop = [img.get_rect().centerx, 32 * i])
        img.blit(limg, r)
    return img



# charselect.py -- character selection screen
# Copyright 2004 Joe Wreschnig <piman@sacredchao.net>
# Released under the terms of the GNU GPL v2.
__revision__ = "$Id: charselect.py 286 2004-09-04 03:51:59Z piman $"

import pygame
from pygame.sprite import Sprite, RenderUpdates

import textfx
import wipes
import load

from events import EventManager
from characters import Character

from constants import *

# Display the character selection screen; return the two selected
# characters (or None, None when escape was pressed).
def init(numplayers = 2):
    screen = pygame.display.get_surface()
    em = EventManager()
    bg = load.image("select-bg.png")
    move_snd = load.sound("select-move.wav")
    confirm_snd = load.sound("select-confirm.wav")
    confirm_snd.set_volume(0.4)

    sprites = RenderUpdates()
    portraits = [PortraitDisplay([20, 20]), PortraitDisplay([20, 320])]
    names = [NameDisplay([20, 220]), NameDisplay([20, 520])]
    drops = [DropDisplay([240, 20]), DropDisplay([240, 320])]
    stats = [StatDisplay([240, 150]), StatDisplay([240, 450])]
    descs = [DescDisplay([430, 20]), DescDisplay([430, 320])]
    char_sprites = list(zip(portraits, names, drops, stats, descs))

    idx = [0, 0]
    confirmed = [False, False]
    for i, sprs in enumerate(char_sprites):
        if i < numplayers:
            for spr in sprs: spr.set_char(Character.available[i])
            sprites.add(sprs)
            idx[i] = i

    init_bg = bg.convert()
    sprites.update(pygame.time.get_ticks())
    sprites.draw(init_bg)
    wipes.wipe_in(init_bg)
    init_bg = None # Let us GC it
    pygame.display.update()
    sprites.clear(screen, bg)

    while False in [(c.confirmed or i >= numplayers)
                    for i, c in enumerate(portraits)]:
        for ev in em.wait():
            if ev.type == PLAYER:
                if ev.key == LEFT:
                    i = (idx[ev.player] - 1) % len(Character.available)
                    idx[ev.player] = i
                elif ev.key == RIGHT:
                    i = (idx[ev.player] + 1) % len(Character.available)
                    idx[ev.player] = i
                elif ev.key in [ROT_CC, ROT_CW, CONFIRM]:
                    confirm_snd.play()
                    portraits[ev.player].confirmed = True

                if ev.key in [LEFT, RIGHT]:
                    move_snd.play()
                    for spr in char_sprites[ev.player]:
                        spr.set_char(Character.available[idx[ev.player]])
                    portraits[ev.player].confirmed = False

            elif ev.type == QUIT:
                return None, None

        sprites.update(pygame.time.get_ticks())
        pygame.display.update(sprites.draw(screen))
        sprites.clear(screen, bg)

    return [Character.available[i] for i in idx]

# A character portrait, 200x200 image.
class PortraitDisplay(Sprite):
    def __init__(self, topleft):
        Sprite.__init__(self)
        self.rect = Rect([topleft, [200, 200]])
        self._light = pygame.Surface([180, 180])
        self._light.fill([255, 255, 255])
        self._dark = pygame.Surface([180, 180])
        self._dark.fill([0, 0, 0])
        self.confirmed = False

    def set_char(self, char):
        charimage = char.images["portrait"].convert()
        charimage.set_alpha(130)
        self._dark = pygame.Surface([180, 180])
        self._dark.fill([0, 0, 0])
        self._dark.blit(charimage, [0, 0])
        self._dark = char.border(self._dark)

        charimage.set_alpha(256)
        self._light = pygame.Surface([180, 180])
        self._light.blit(charimage, [0, 0])
        self._light = char.border(self._light)

    def update(self, time):
        if self.confirmed: self.image = self._light
        else: self.image = self._dark

# The character's name
class NameDisplay(Sprite):
    def __init__(self, topleft):
        Sprite.__init__(self)
        self._topleft = topleft

    def set_char(self, char):
        self.image = textfx.shadow(char.name, 50)
        self.rect = self.image.get_rect(topleft = self._topleft)

# A display of the character's drop gem pattern.
class DropDisplay(Sprite):
    def __init__(self, topleft):
        Sprite.__init__(self)
        self._topleft = topleft

    def set_char(self, char):
        self.image = char.border(char.drop.render())
        self.rect = self.image.get_rect(topleft = self._topleft)

# Display statistics about the character.
class StatDisplay(Sprite):
    def __init__(self, topleft):
        Sprite.__init__(self)
        self._topleft = topleft
        self._fsize = 30
        self._height = pygame.font.Font(None, self._fsize).get_linesize()

    def set_char(self, char):
        stats = char.stats
        y = self._height * len(stats) + self._height//2
        self.image = pygame.Surface([400, y], SRCALPHA, 32)

        for i, stat in enumerate(stats):
            self.image.blit(textfx.shadow(stat, self._fsize),
                            [0, i * self._height])
        self.rect = self.image.get_rect(topleft = self._topleft)

# A description of the character. If the description is too long, it
# overflows off the screen...
class DescDisplay(Sprite):
    def __init__(self, topleft):
        Sprite.__init__(self)
        self._topleft = topleft
        self._font = textfx.WrapFont(18, 310)

    def set_char(self, char):
        text = self._font.render(char.description)
        self.image = char.border([text.get_width() + 20,
                                  text.get_height() + 20])
        self.image.blit(text, [20, 20])
        self.rect = self.image.get_rect(topleft = self._topleft)



# config.py -- ConfigFile proxy/wrapper module
# Copyright 2004 Joe Wreschnig <piman@sacredchao.net>
# Released under the terms of the GNU GPL v2.
__revision__ = "$Id: config.py 291 2004-09-09 00:38:52Z piman $"

import os
from configparser import ConfigParser
from constants import *

_config = ConfigParser()

# Alias various ConfigParser functions.
getboolean = _config.getboolean
getint = _config.getint
set = _config.set
get = _config.get

_unlocked = {}

# Set up default options and high scores, then load the real
# config file (if it exists).
def init(rc_file):
    _config.add_section("settings")
    _config.set("settings", "rotate_on_up", "yes")
    _config.set("settings", "rotate_on_space", "yes")
    _config.set("settings", "matches", "3")
    _config.set("settings", "ai", "2")
    _config.set("settings", "speed", "750")
    _config.set("settings", "combat", "no")
    _config.set("settings", "fullscreen", "yes")

    _config.add_section("buttons")
    _config.set("buttons", "start", "9")
    _config.set("buttons", "pause", "8")

    _config.add_section("unlock")
    for c in ["unlocker", "unixbros", "combat", "single", "cpuversus"]:
        _config.set("unlock", c, "no")

    _config.add_section("scores-single")
    for n, name, score in [(1, "Ogi", 1000),
                           (2, "Gnarr", 500),
                           (3, "Thane", 0)]:
        _config.set("scores-single", str(n), "%s,%d" % (name, score))

    _config.add_section("scores-versus-arcade")
    for n, name, score in [(1, "Ogi", 1000),
                           (2, "Gnarr", 500),
                           (3, "Thane", 0)]:
        _config.set("scores-versus-arcade", str(n), "%s,%d" % (name, score))

    for c in range(1, 10, 2):
        _config.add_section("scores-versus-%d" % c)
        for n, name, score in [(1, "Ogi", 1000),
                               (2, "Gnarr", 500),
                               (3, "Thane", 0)]:
            _config.set("scores-versus-%d" % c, str(n),
                       "%s,%d" % (name, score))

    if os.name == "posix": rc_file = os.environ["HOME"] + "/.angryddrc"
    else: rc_file = "angryddrc"
    _config.read([rc_file])

    if _config.getboolean("settings", "rotate_on_space") is False:
        from events import EventManager
        EventManager().events[K_SPACE] = DOWN

# Unlock things, but delay until game quit.
def unlock(thing): _unlocked[thing] = True

def write(*args):
    for key in _unlocked: _config.set("unlock", key, "yes")
    _config.write(*args)

# Set various options "intelligently". Up/down change the option,
# Confirm increases it but wraps from the highest value to the lowest.

def set_matches(menu, platform, pos, key):
    m = _config.getint("settings", "matches")
    if key == CONFIRM: m = (m + 2) % 10
    elif key == UP or key == ROT_CW: m = min(9, m + 2)
    else: m = max(1, m - 2)
    _config.set("settings", "matches", str(m))
    platform.text = get_matches()
    return False

def get_matches():
    m = _config.getint("settings", "matches")
    return "Best %d/%d" % ((m // 2) + 1, m)

def set_speed(menu, platform, pos, key):
    s = _config.getint("settings", "speed")
    if key == CONFIRM:
        if s == 250: s = 1250
        else: s -= 250
    elif key == UP or key == ROT_CW: s = max(250, s - 250)
    else: s = min(1250, s + 250)
    _config.set("settings", "speed", str(s))
    platform.text = get_speed()
    return False

def get_speed():
    return { 250: "Very Fast", 500: "Fast", 750: "Normal Speed",
             1000: "Slow", 1250: "Very Slow"
             }[_config.getint("settings", "speed")]

def set_rotup(menu, platform, pos, key):
    r = _config.getboolean("settings", "rotate_on_up")
    r ^= True
    _config.set("settings", "rotate_on_up", str(r).lower())
    platform.text = get_rotup()
    return False

def get_rotup():
    if _config.getboolean("settings", "rotate_on_up"): return "Up Rotates"
    else: return "Up Doesn't Rotate"

def set_space(menu, platform, pos, key):
    r = _config.getboolean("settings", "rotate_on_space")
    r ^= True
    _config.set("settings", "rotate_on_space", str(r).lower())
    from events import EventManager
    if r: EventManager().events[K_SPACE] = ROT_CC
    else: EventManager().events[K_SPACE] = DOWN
    platform.text = get_space()
    return False

def get_space():
    if _config.getboolean("settings", "rotate_on_space"):
        return "Space Rotates"
    else: return "Space Drops"

def set_combat(menu, platform, pos, key):
    r = _config.getboolean("settings", "combat")
    r ^= True
    _config.set("settings", "combat", str(r).lower())
    platform.text = get_combat()
    return False

def get_combat():
    if _config.getboolean("settings", "combat"): return "Combat Blocks"
    else: return "No Combat Blocks"

def set_ai(menu, platform, pos, key):
    m = _config.getint("settings", "ai")
    if key == CONFIRM: m = (m + 1) % 6
    elif key == UP or key == ROT_CW: m = min(5, m + 1)
    else: m = max(0, m - 1)
    _config.set("settings", "ai", str(m))
    platform.text = get_ai()
    return False

def get_ai():
    return "AI: %s" % ["Stupid", "Very Easy", "Easy", "Normal",
                       "Hard", "Insane"][_config.getint("settings", "ai")]



# events.py -- event handling wrapper
# Copyright 2004 Joe Wreschnig <piman@sacredchao.net>
# Released under the terms of the GNU GPL v2.
__revision__ = "$Id: events.py 317 2006-01-12 21:19:27Z piman $"

from constants import *
from pygame import event, key, joystick, time, display
from pygame.event import Event

from util import Singleton

# Key repeat delay and repeat rate and joystick tolerance.
DELAY, RATE = 200, 75
TOLERANCE = 0.95

import config

CONFIRM_BUTTON = config.getint("buttons", "start")
PAUSE_BUTTON = config.getint("buttons", "pause")

# Alias post, so pygame.event is unnecessary outside this module.
post = event.post

class EventManager(Singleton):
    def __init__(self):
        if hasattr(EventManager, "_singleton"): return
        key.set_repeat(DELAY, RATE)
        event.set_allowed(None)
        event.set_allowed([KEYDOWN, JOYBUTTONDOWN, JOYAXISMOTION,
                           JOYBUTTONUP, PLAYER])

        for i in range(joystick.get_count()):
            joystick.Joystick(i).init()

        # Information for joystick repeat.
        self._last_press = [None, None]
        self._last_press_start = [0, 0]
        self._last_press_times = [0, 0]

        # Maps key values to player numbers.
        self.players = { K_w: 0, K_a: 0, K_s: 0, K_d: 0, K_q: 0, K_e: 0,
                         K_i: 1, K_j: 1, K_k: 1, K_l: 1, K_u: 1, K_o: 1,
                         K_LEFT: 0, K_RIGHT: 0, K_UP: 0, K_DOWN: 0,
                         K_SPACE: 0, K_KP_ENTER: 0, K_RETURN: 0, K_2: 1,
                         K_1: 0, K_BACKSPACE: 1 }

        # Maps key values to internal event keys.
        self.events = { K_w: UP, K_a: LEFT, K_s: DOWN, K_d: RIGHT,
                        K_q: ROT_CC, K_e: ROT_CW,
                        K_i: UP, K_j: LEFT, K_k: DOWN, K_l: RIGHT,
                        K_u: ROT_CC, K_o: ROT_CW,
                        K_LEFT: LEFT, K_RIGHT: RIGHT, K_UP: UP, K_DOWN: DOWN,
                        K_SPACE: ROT_CC, K_KP_ENTER: CONFIRM,
                        K_RETURN: CONFIRM, K_2: CONFIRM,
                        K_BACKSPACE: CONFIRM, K_1: CONFIRM }

        Singleton.__init__(self)

    def clear(self):
        self._last_press = [None, None]
        self._last_press_start = [0, 0]
        self._last_press_times = [0, 0]
        event.clear()

    # This isn't a very good "wait" function.
    def wait(self):
        while not event.peek([KEYDOWN, JOYBUTTONDOWN, JOYAXISMOTION,
                              JOYBUTTONUP, QUIT]):
            time.wait(10)
        return self.get()

    # Get all events in the Pygame event queue and translate them
    # to PLAYER events.
    def get(self):
        t = time.get_ticks()
        events = []
        for ev in event.get(KEYDOWN):
            try:
                events.append(Event(PLAYER,
                                    key = self.events[ev.key],
                                    player = self.players[ev.key]))
            except KeyError:
                if ev.key == K_F11 or ev.key == K_f:
                    display.toggle_fullscreen()
                elif ev.key == K_p:
                    events.append(Event(PAUSE))
                elif ev.key == K_ESCAPE:
                    events.append(Event(QUIT))

        # Joysticks are translated to either 0 or 1, and any button
        # below 4 is also translated to 0 or 1. Thus, e.g. button 3
        # on js 1 and button 1 on js 3 will generate the same event.

        # Button numbers over 15 swap the joystick value and have
        # 16 subtracted from them. This makes the EMSUSB2 adapter work
        # in Linux (where it appears as a single 32 button joystick,
        # rather than two 16 button ones).

        # Button patterns testing on Playstation controllers and also
        # a Gravis gamepad.
        for ev in event.get(JOYBUTTONDOWN):
            if ev.button > 15: button, joy = ev.button - 16, ev.joy + 1
            else: button, joy = ev.button, ev.joy
            player = joy % 2
            if button == CONFIRM_BUTTON:
                events.append(Event(PLAYER, key = CONFIRM, player = player))
            elif button == PAUSE_BUTTON:
                events.append(Event(PAUSE))
            elif button < 4:
                if button & 1: key = ROT_CW
                else: key = ROT_CC
                self._last_press[player] = key
                self._last_press_start[player] = t
                events.append(Event(PLAYER, key = key, player = player))

        for ev in event.get(JOYBUTTONUP):
            if ev.button > 15: button, joy = ev.button - 16, ev.joy + 1
            else: button, joy = ev.button, ev.joy
            player = joy % 2
            if button < 4:
                if button & 1: key = ROT_CW
                else: key = ROT_CC
                if self._last_press[player] == key:
                    self._last_press[player] = None

        # Axis stuff; again for the EMSUSB2 axes 3 and 4 are actually on
        # the second joystick.
        for ev in event.get(JOYAXISMOTION):
            if ev.axis == 3: axis, joy = 0, ev.joy + 1
            elif ev.axis == 4: axis, joy = 1, ev.joy + 1
            else: axis, joy = ev.axis, ev.joy
            player = joy % 2
            key = None
            if axis == 0:
                if ev.value < -TOLERANCE: key = LEFT
                elif ev.value > TOLERANCE: key = RIGHT
            elif axis == 1:
                if ev.value > TOLERANCE: key = DOWN
                elif ev.value < -TOLERANCE: key = UP

            if key is not None and self._last_press[player] != key:
                self._last_press[player] = key
                self._last_press_start[player] = t
                events.append(Event(PLAYER, key = key, player = player))
            elif key is None and axis <= 1:
                self._last_press[player] = None

        # Check for repeated events -- if any player has a pending
        # repeat event that hasn't fired in the given interval put it
        # on the queue again.
        for player in range(len(self._last_press)):
            if (self._last_press[player] is not None and
                (t - self._last_press_start[player]) > DELAY and
                (t - self._last_press_times[player]) > RATE):
                self._last_press_times[player] = t
                events.append(Event(PLAYER, key = self._last_press[player],
                                    player = player))

        # Don't actually post events here; the loop that gets them
        # knows what to do. However, if PLAYER events are posted, grab them
        # and use them.
        events.extend(event.get([PLAYER, QUIT]))
        return events



# ai.py -- AI and helper routines for Angry, Drunken Dwarves
# Copyright 2004 Joe Wreschnig <piman@sacredchao.net>
# Released under the terms of the GNU GPL v2.
__revision__ = "$Id: ai.py 286 2004-09-04 03:51:59Z piman $"

# Trick borrowed from Peter Norvig's Python IAQ:
# http://www.norvig.com/python-iaq.html
def _abstract():
    import inspect
    caller = inspect.getouterframes(inspect.currentframe())[1][3]
    raise NotImplementedError(caller + ' must be implemented in subclass')

import random, math

import util
from game import BasicField
from boxes import TickBox, BreakBox, Box, Special, Diamond

def copy_box(box):
    args = [box.color, [box.x, box.y]]
    newbox = None
    if isinstance(box, Box):
        newbox = Box(*args)
    elif isinstance(box, TickBox):
        newbox = TickBox(*args)
        newbox.time_left = box.time_left
    elif isinstance(box, BreakBox):
        newbox = BreakBox(*args)
    elif isinstance(box, Special):
        newbox = Special(box.color, [box.x, box.y], box.special)
    elif isinstance(box, Diamond):
        newbox = Diamond(box.color, [box.x, box.y])
    return newbox

class AIField(BasicField):
    def __init__(self, field):
        BasicField.__init__(self, field.width, field.height)
        for y, row in enumerate(field):
            for x, box in enumerate(row):
                if field[y][x] is not None:
                    # Only add a box if we're at its topleft, i.e.
                    # don't add merged boxes many times.
                    newbox = copy_box(box)
                    if isinstance(newbox, Box):
                        if box.x != x or box.y != y: continue
                        else: newbox.set_size(box.size, self)
                    if newbox:
                        self.add_box([newbox.x, newbox.y], newbox)

    # Take a turn (move all pieces down, tick, merge, break, repeat)
    # Doesn't do boning. Returns the value of gems broken.
    def take_turn(self):
        while self.fall(): pass
        self.tick()
        while self.merge(): pass
        value, dead = self.breaking()
        while len(dead) > 0:
            while self.fall(): pass
            while self.merge(): pass
            newv, dead = self.breaking()
            value += newv
        return value

    def can_drop1(self, col):
        return (self._field[0][col] is None)

    # Orientation is as described in FallingBoxes#_render.
    def can_drop(self, col, orientation):
        if col >= self.width: return False
        elif orientation == 0: # 2 above 1
            return (self._field[0][col] is None)
        elif orientation == 1:
            return ((col != 0) and self._field[0][col] is None and
                    self._field[0][col - 1] is None)
        elif orientation == 2:
            return (self._field[0][col] is None and
                    self._field[1][col] is None)
        elif orientation == 3:
            return (col != self.width - 1 and
                    self._field[0][col] is None and
                    self._field[0][col + 1] is None)
        else: return False

    # Drop box1 and box2 in the given orientation at the top of this
    # field. Return True (and does it if it thinks such a thing is
    # possible, False otherwise. Possibility is currently only
    # decided by space available at the top, and not any intervening
    # blocks that might block it (i.e. naively)
    def drop(self, boxes, col, orientation):
        box1, box2 = boxes
        if not self.can_drop(col, orientation): return False
        
        elif orientation == 0: # 2 above 1
            if self._field[1][col] is not None:
                self.add_box([col, 0], box1)
            else:
                self.add_box([col, 1], box1)
                self.add_box([col, 0], box2)

        elif orientation == 1: # 2 left of 1
            self.add_box([col, 0], box1)
            self.add_box([col - 1, 0], box2)

        elif orientation == 2: # 2 below 1
            self.add_box([col, 0], box1)
            self.add_box([col, 1], box2)

        elif orientation == 3: # 2 right of 1
            self.add_box([col, 0], box1)
            self.add_box([col + 1, 0], box2)

        return True

    def drop1(self, box, col):
        if not self.can_drop1(col): return False
        else:
            self.add_box([col, 0], box)
            return True

    def get_height(self, col):
        for y in range(self.height):
            if self._field[y][col] != None: return self.height - y
        else: return 0

    def max_height(self):
        return max([self.get_height(i) for i in range(self.width)])

    def average_height(self):
        s = 4 * self.get_height(3) # This is very bad to fill up
        s += sum([self.get_height(x) for x in range(self.width)])
        return s // self.width

    def stddev_height(self):
        avg = self.average_height()
        std_dev = 0
        for i in range(self.width):
            v = avg - self.get_height(i)
            std_dev += v * v
        std_dev /= float(self.width)
        return math.sqrt(std_dev)

    # Expand given only one box; total states = field.width.
    def expand1(self, box, rand = True):
        columns = list(range(self.width))
        if rand: random.shuffle(columns)

        for col in columns:
            if self.can_drop1(col):
                f = AIField(self)
                newbox = copy_box(box)
                f.drop1(newbox, col)
                yield 0, col, f, [newbox]

    # Expand the state space based on the boxes given. 4 orientations,
    # width of 6 = 22 expansions (after illegal ones removed).
    def expand(self, boxes, rand = True):
        orientations = list(range(4))
        columns = list(range(self.width))
        if rand:
            random.shuffle(orientations)
            random.shuffle(columns)
            columns.remove(3); columns.append(3)

        for orientation in orientations:
            for col in columns:
                if self.can_drop(col, orientation):
                    f = AIField(self)
                    newboxes = [copy_box(b) for b in boxes]
                    f.drop(newboxes, col, orientation)
                    yield orientation, col, f, newboxes

class AI(object):
    def __init__(self, player):
        self.player = player
        self.insane = False
        self.drops = 70
        self.delta = 120

    # Expand and find the highest value for the given callback; it
    # defaults to the value of the gems broken if no callback is
    # passed in.
    def find_high(self, h = (lambda *args: args[1])):
        best_score = None
        best_move = (0, 0)
        for orient, col, field, boxes in self.field.expand(self.falling):
            fscore = field.take_turn()
            score = h(orient, col, field, boxes, fscore)
            if best_score is None or score > best_score:
                best_score = score
                best_move = (orient, col)
            yield None
        yield best_move
        return

    # Expand and find the lowest value for the given callback; if
    # the value is 0, it returns it and doesn't expand anymore.
    def find_low(self, h, halt_on_zero = True):
        best_score = None
        best_move = (0, 0)
        for orient, col, field, boxes in self.field.expand(self.falling):
            fscore = field.take_turn()
            score = h(orient, col, field, boxes, fscore)
            if best_score is None or score < best_score:
                best_score = score
                best_move = (orient, col)
            if best_score == 0 and halt_on_zero: break
            yield None
        yield best_move
        return

    def incoming(self, field, falling):
        self.field = AIField(field)
        self.falling = [falling._box1, falling._box2]

    def __iter__(self):
        # Override this function to change how the AI moves. It should
        # return an iterator (or generator) that returns either None
        # when it's still thinking, or a tuple (orientation, column)
        # to put the piece in when it's done. This function is called
        # when a new piece is put into play, and after a move is made
        # is not called again until another piece is put in. (i.e.
        # you can't change your mind.)

        # Note that once your moves are committed, each left/right/rotate
        # motion will take a (nontrivial) fraction of a second! Around
        # 1/12th of a second per move. So, thinking fast is important.

        # Look at the RandomAI for a trivial example.
        _abstract()

class SlowAI(AI):
    def __init__(self, *args):
        AI.__init__(self, *args)
        self.drops = 0
        self.delta = 500

class MediumAI(AI):
    def __init__(self, *args):
        AI.__init__(self, *args)
        self.drops = 100
        self.delta = 150

class FastAI(AI):
    def __init__(self, *args):
        AI.__init__(self, *args)
        self.drops = 80
        self.delta = 100

class SuperFastAI(AI):
    def __init__(self, *args):
        AI.__init__(self, *args)
        self.drops = 25
        self.delta = 25

# A test AI that plays totally randomly.
class RandomAI(AI):
    def __iter__(self):
        yield None
        yield (random.randrange(0, 4),
               random.randrange(0, self.player.field.width))

# Another test AI, just tries to keep average field height low
class KeepLowAI(AI):
    def __iter__(self):
        return self.find_low(self._total_height)

    def _total_height(self, orient, col, field, boxes, score):
        return sum([field.get_height(x) for x in range(field.width)])

class GoodAI(AI):
    def __iter__(self):
        if self.field.get_height(3) > 8:
            return self.find_low(self._emergency)
        elif self.field.max_height() > 9:
            return self.find_high(self._lower_max)
        elif self.field.average_height() < 7:
            return self.find_high(self._form_boxes)
        else:
            return self.find_high()

    def _emergency(self, orient, col, field, boxes, score):
        return (score > 25, field.get_height(3), score)

    def _lower_max(self, orient, col, field, boxes, score):
        return (
            score > 25,
            -field.max_height(),
            score,
            self._gemsize(field))

    def _form_boxes(self, orient, col, field, boxes, score):
        return (
            score > 25,
            field.max_height() < 10,
            self._gemsize(field),
            self._adjacencies(field, boxes))

    def _adjacencies(self, field, boxes):
        adj = 0
        for box in filter(None, boxes):
            if not box.crashed:
                for adjbox in filter(None, box.adjacent(field)):
                    if adjbox.color == box.color: adj += 3
                    else: adj -= 1
        return adj

    def _gemsize(self, field):
        return max([1] + [box.size[0] * box.size[1]
                          for box in util.flatten(field)
                          if box])

class StupidAI(RandomAI, SlowAI): pass
class VeryEasyAI(KeepLowAI, SlowAI): pass
class EasyAI(KeepLowAI, MediumAI): pass
class NormalAI(GoodAI, MediumAI): pass
class HardAI(GoodAI, FastAI): pass
class InsaneAI(GoodAI, SuperFastAI): pass



# load.py -- sound/image loading wrappers for graceful failure
# Copyright 2004 Joe Wreschnig <piman@sacredchao.net>
# Released under the terms of the GNU GPL v2.
__revision__ = "$Id: load.py 296 2005-10-06 06:36:15Z piman $"

import pygame
import pygame_ext
import math

from dirstore import DirStore
from constants import *

color_cache = {}

HUE_MAP = {
    "red": 0,
    "orange": 25,
    "yellow": 60,
    "green": 120,
    "cyan": 175,
    "blue": 230,
    "purple": 295,
    }

GEM_POLYGON = [
    # the overall shape, so there are no gaps:
    ([1.00, 0.50],
     [[0, 4], [4,0], [28, 0], [32, 4], [32, 28], [28, 32], [4, 32], [4, 28]]),
    # individual facets, top to bottom, right to left:
    # top row
    ([0.20, 0.90], [[0, 4], [4, 0], [16, 0]]),
    ([0.60, 0.70], [[4, 0], [16, 0], [16, 4]]),
    ([0.50, 0.75], [[16, 0], [28, 0], [16, 4]]),
    ([1.00, 0.50], [[0, 4], [4, 0], [16, 4]]),
    # second row
    ([0.55, 0.75], [[0, 4], [4, 4], [4, 16], [0, 16]]),
    ([0.85, 0.55], [[4, 4], [16, 4], [16, 16], [4, 16]]),
    ([1.00, 0.35], [[16, 4], [28, 4], [28, 16], [16, 16]]),
    ([0.75, 0.60], [[28, 4], [32, 4], [32, 16], [28, 16]]),
    # third row
    ([0.85, 0.55], [[0, 16], [4, 16], [4, 28]]),
    ([1.00, 0.25], [[0, 16], [4, 28], [0, 28]]),
    ([1.00, 0.40], [[4, 16], [16, 16], [16, 28], [4, 28]]),
    ([1.00, 0.10], [[16, 16], [28, 16], [28, 28], [16, 28]]),
    ([0.65, 0.70], [[28, 16], [32, 16], [32, 28]]),
    ([0.35, 0.85], [[28, 16], [32, 28], [28, 28]]),
    # bottom row
    ([0.85, 0.60], [[0, 28], [8, 28], [4, 32]]),
    ([1.00, 0.10], [[4, 32], [8, 28], [16, 28], [16, 32]]),
    ([0.65, 0.70], [[16, 28], [28, 32], [16, 32]]),
    ([0.40, 0.80], [[16, 28], [28, 28], [28, 32]]),
    ([0.20, 0.90], [[28, 28], [32, 28], [28, 32]])
    ]

# Load an image of a block of the given color and type. From disk
# if this is the first time loading it; otherwise, from a dict cache.
def block(color, typ = ""):
    if (color, typ) in color_cache:
        return color_cache[(color, typ)].copy()
    else:
        img = image(color + typ + ".png")
        color_cache[(color, typ)] = img.convert()
        return img.copy()

def hsv_to_rgb(h,s,v):
    rgb = []
    if s == 0: rgb = [v, v, v] # grayscale
    elif v == 1: rgb = [1, 1, 1] # white
    elif v == 0: rgb = [0, 0, 0] # black
    else:
        h = h / 60.0
        i = math.floor(h)
        f = h - i
        p = v * (1 - s)
        q = v * (1 - s * f)
        t = v * (1 - s * (1 - f))
        if   i == 0: rgb = [v, t, p]
        elif i == 1: rgb = [q, v, p]
        elif i == 2: rgb = [p, v, t]
        elif i == 3: rgb = [p, q, v]
        elif i == 4: rgb = [t, p, v]
        else: rgb = [v, p, q]
    for i in range(3):
        rgb[i] = int(rgb[i] * 255)
    return rgb

def gem(color, w, h):
    if (color, w, h) in color_cache:
        return color_cache[(color, w, h)].copy()

    if color == "gray": hue = 0
    else: hue = HUE_MAP[color]

    surf = pygame.Surface([w * 32, h * 32], 0, 32)
    for entry in GEM_POLYGON:
        if color == "gray": col = hsv_to_rgb(hue, 0, entry[0][1])
        else: col = hsv_to_rgb(hue, entry[0][0], entry[0][1])
        points = []
        if (w == 1 and h == 1):
            points = entry[1]
        else:
            for point in entry[1]:
                points += [[point[0] * w, point[1] * h]]
        pygame.draw.polygon(surf, col, points)
        pygame.draw.aalines(surf, col, True, points, True)
    surf = surf.convert()
    surf.set_colorkey(surf.get_at([0, 0]))
    color_cache[(color, w, h)] = surf.copy()
    return surf

# Load a sound. pygame_ext makes this into a fake object if necessary.
def sound(filename):
    path = os.path.join(angrydd_path, "sounds")
    if os.path.exists(path + ".zip"): store = DirStore(path + ".zip")
    else: store = DirStore(path)
    try: return pygame.mixer.Sound(store.open(filename))
    except ValueError as s:
        #print "W: Unable to load %s: %s" % (filename, s)
        return pygame_ext.FakeSound()
# Load an image, or exit gracefully if it fails.
def image(filename):
    path = os.path.join(angrydd_path, "images")
    if os.path.exists(path + ".zip"): store = DirStore(path + ".zip")
    else: store = DirStore(path)

    try: return pygame.image.load(store.open(filename))
    except pygame.error:
        raise SystemExit("E: Unable to load %s %s" % (filename, err))



# characters.py -- character loading and Drop Pattern display/generation
# Copyright 2004 Joe Wreschnig <piman@sacredchao.net>
# Released under the terms of the GNU GPL v2.
__revision__ = "$Id: characters.py 286 2004-09-04 03:51:59Z piman $"

import os
import math
import pygame
from pygame import transform

import dirstore; dirstore.require("0.1")
from dirstore import DirStore

from functools import cmp_to_key

import config
import random
import load

from constants import *

# Stores information about a character (not a player).
class Character(object):
    def __init__(self, name):
        store = DirStore(name)
        self.drop = DropPattern([s.strip() for s in
                                 store.read("drop").strip().split("\n")])
        self.pattern = self.drop.render()
        self.stats = store.read("description").strip().split("\n")
        if "quips" in store:
            self.quips = store.read("quips").strip().split("\n")
        else:
            self.quips = Character.default.quips
        self.description = self.stats.pop()
        self.name = self.stats.pop(0).strip()
        self.stats.append("Attack: %0.2fx" % self.drop.cost)
        self.images = {}

        for image in ["bl-corner", "bottom-bar", "br-corner", "left-bar",
                      "panel-bg", "portrait", "right-bar", "top-bar",
                      "ul-corner", "ur-corner", "background"]:
            filename = image + ".png"
            if filename in store:
                self.images[image] = pygame.image.load(store.open(filename))
            else:
                self.images[image] = Character.default.images[image]

    # Draw a border, optionally with an image in the center
    def border(self, image):
        # Passing in a sequence fills the image with the character's
        # background.
        if isinstance(image, list) or isinstance(image, tuple):
            img = self.images["panel-bg"]
            surf = pygame.Surface([image[0] + 20, image[1] + 20])
            for x in range(0, image[0], img.get_width()):
                for y in range(0, image[1], img.get_height()):
                    surf.blit(img, [x + 10, y + 10])

        else:
            surf = pygame.Surface([image.get_width() + 20,
                                   image.get_height() + 20])
            surf.blit(image, [10, 10])

        top = self.images["top-bar"]
        bot = self.images["bottom-bar"]
        bot_y = surf.get_height() - 10
        for x in range(0, surf.get_width(), top.get_width()):
            surf.blit(top, [x, 0])
        for x in range(0, surf.get_width(), bot.get_width()):
            surf.blit(bot, [x, bot_y])

        left = self.images["left-bar"]
        right = self.images["right-bar"]
        right_x = surf.get_width() - 10
        for y in range(0, surf.get_height(), left.get_height()):
            surf.blit(left, [0, y])
        for y in range(0, surf.get_height(), right.get_height()):
            surf.blit(right, [right_x, y])

        surf.blit(self.images["ul-corner"], [0, 0])
        surf.blit(self.images["ur-corner"], [right_x, 0])
        surf.blit(self.images["bl-corner"], [0, bot_y])
        surf.blit(self.images["br-corner"], [right_x, bot_y])

        return surf

# The pattern which a character drops gems in.
class DropPattern(object):
    color_map = { "b": "blue", "g": "green", "y": "yellow", "r": "red",
                  "p": "purple", "c": "cyan", "o": "orange" }

    def __init__(self, array):
        self._array = [[DropPattern.color_map[c] for c in s] for s in array]
        self.cost = self._calculate_cost()

    # Given a number of gems to drop, return the actual rows of blocks
    # to be dropped.
    def multiply(self, count):
        rows = []
        array = list(self._array)
        if self.cost < 1: count = int(math.ceil(count * self.cost))
        elif self.cost > 1: count = int(math.floor(count * self.cost))
        while count >= 6:
            rows.append(list(array.pop()))
            array.insert(0, list(rows[-1]))
            count -= 6
        if count != 0:
            possible = [0, 1, 2, 4, 5]
            to_fill = []
            rows.append(list(array.pop()))
            array.insert(0, list(rows[-1]))
            while count > 0 and len(possible) > 0:
                to_fill.append(random.choice(possible))
                possible.remove(to_fill[-1])
                count -= 1

            # Column 3 (where pieces fall) is filled in last
            if count == 1: to_fill.append(3)

            for i in range(len(rows[-1])):
                if i not in to_fill: rows[-1][i] = None
            
        return rows

    # Draw our gem pattern as an image, for the character selection screen.
    def render(self):
        pattern = pygame.Surface([32 * 6, 32 * 4])

        for y, row in enumerate(self._array):
            for x, c in enumerate(row):
                img = load.block(c)
                pattern.blit(img, [32 * x, 32 * y])
        pattern = transform.rotozoom(pattern, 0, 0.8).convert()
        return pattern

    def _calculate_cost(self):
        # Patterns that are more complicated drop less stones, to be
        # fair. Thus, we have to calculate the "randomness" of the drop
        # pattern, and get some kind of multiplier to make it fair. So:
        #
        # For each stone, look at the stones to the left, and below
        # it. If both are the same color, add 1. If neither are the
        # same color, subtract 1. If one is the same and one is
        # different, add 0. This gives us an initial score; higher
        # means the grid is "easier", and so the multiplier should be
        # higher.
        #
        # Thus, -20 is the "worst" (.5x) and 20 is the "best" (2x).
        
        s = 0
        for i in range(4):
            for j in range(5):
                c = self._array[i][j]
                c1 = self._array[(i + 1) % 4][j]
                c2 = self._array[i][j + 1]

                if c1 == c2 == c: s += 1
                elif c1 != c and c2 != c: s -= 1

        if s < 0: return 1 - s / -40.0
        else: return 1 + (s / 20.0)

# Load all characters found in our resource paths
def init():
    char_base = os.path.join(angrydd_path, "characters")
    try:
        Character.default = Character(os.path.join(char_base, "default"))
    except StandardError as s:
        try:
            Character.default = Character(os.path.join(char_base,
                                                       "default.zip"))
        except:
            raise SystemExit("E: Unable to load default character data.")

    dir = os.listdir(char_base)

    avail = []
    arcade = []

    # Remove the secret characters if they aren't unlocked yet.
    if not config.getboolean("unlock", "unixbros"):
        for name in ["cron", "grep"]:
            try: dir.remove(name + ".dwarf")
            except ValueError: pass
            try: dir.remove(name + ".dwarf.zip")
            except ValueError: pass

    for name in ["beardy", "ogi", "thane", "rorick", "tharl",
                 "bodger", "gnarr", "cron"]:
        try:
            name += ".dwarf"
            arcade.append(Character(os.path.join(char_base, name)))
        except StandardError as s:
            try:
                name += ".zip"
                arcade.append(Character(os.path.join(char_base, name)))
            except StandardError as s:
                print("W: Unable to load character %s: %s" % (name, s))
                print("W: Arcade mode disabled.")

    for name in dir:
        if "dwarf" in name:
            try:
                avail.append(Character(os.path.join(char_base, name)))
            except StandardError as s:
                print("W: Unable to load character %s: %s" % (name, s))
    avail.sort(key=cmp_to_key(lambda a, b: (a.name > b.name) - (a.name < b.name)))
    Character.available = avail
    Character.arcade = arcade



# boxes.py -- sundry boxes and box-like things for the game
# Copyright 2004 Joe Wreschnig <piman@sacredchao.net>
# Released under the terms of the GNU GPL v2.
__revision__ = "$Id: boxes.py 286 2004-09-04 03:51:59Z piman $"

import pygame
from pygame import transform
from pygame.sprite import Sprite

import textfx
import random
import config
import load

from constants import *

# Abstract class from which other block types inherit.
class AbstractBox(object):
    def __init__(self, color, topleft):
        self.size = [1, 1]      # How big are we; [1,1] except for Box.
        self.color = color
        self.x, self.y = topleft
        self._ntime = None
        self.crashed = False     # If we've been hit by a crash gem/diamond.

    # Accessors for coordinates. FIXME: Make these properties.
    def _get_top(self): return self.y
    def _get_bottom(self): return self.y + self.size[1]
    def _get_left(self): return self.x
    def _get_right(self): return self.x + self.size[0]

    # Blocks adjacent to it on the top...
    def _adj_top(self, field):
        adj = []
        if self.y != 0:
            for x in range(self.x, self._get_right()):
                adj.append(field[self.y - 1][x])
        return list(filter(None, adj))

    # And so on.
    def _adj_bottom(self, field):
        adj = []
        b = self._get_bottom()
        if b != field.height:
            for x in range(self.x, self._get_right()):
                adj.append(field[b][x])
        return list(filter(None, adj))

    def _adj_left(self, field):
        adj = []
        if self.x != 0:
            for y in range(self.y, self._get_bottom()):
                adj.append(field[y][self.x - 1])
        return list(filter(None, adj))

    def _adj_right(self, field):
        adj = []
        r = self._get_right()
        if r != field.width:
            for y in range(self.y, self._get_bottom()):
                adj.append(field[y][r])
        return list(filter(None, adj))

    # All blocks adjacent to this one.
    def adjacent(self, field):
        adj = self._adj_left(field)
        adj.extend(self._adj_right(field))
        adj.extend(self._adj_top(field))
        adj.extend(self._adj_bottom(field))
        return adj

    # Kill the sprite, and remove it from the field map.
    def remove_from(self, field):
        self.crashed = True
        for x in range(self.x, self._get_right()):
            for y in range(self.y, self._get_bottom()):
                field[y][x] = None

    # Check if something is below us on the field (or we're at the bottom).
    def is_blocked_down(self, field):
        bot = self._get_bottom()
        if bot == field.height: return True
        else:
            for x in range(self.x, self._get_right()):
                if field[bot][x] is not None: return True
            return False

    # Actually move down space.
    def fall(self, field):
        bot = self._get_bottom()
        for x in range(self.x, self._get_right()):
            field[self.y][x] = None
            field[bot][x] = self
        self.y += 1

    # Move to an entirely new location. This is done when blocks are
    # moved from a FallingBlock onto the field.
    def move(self, xy):
        self.x, self.y = xy

    # Destroy self if appropriate; return the value of the gems
    # destroyed.
    def crash(self, field, gem, immed = False):
        broken = 0
        if not self.crashed and gem.color == self.color:
            broken += self.size[0] * self.size[1] * (sum(self.size) // 2)
            self.crashed = True
            for box in self.adjacent(field):
                broken += box.crash(field, self)
            if self.size != [1, 1]:
                field.max_gemsize = max(field.max_gemsize,
                                        self.size[0] * self.size[1])
        return broken

# A generator for random boxes.
class BoxGen(object):
    def __init__(self, rand, colors, single):
        self.rand = rand
        self.combat = (config.getboolean("settings", "combat") and not single)
        self.colors = colors

    def get(self, topleft):
        f = self.rand.random()
        color = self.rand.choice(self.colors)
        if self.combat:
            if f <= 0.004: return DiamondSprite("diamond", topleft)
            elif f <= 0.20: return BreakBoxSprite(color, topleft)
            elif f <= 0.27: return SpecialSprite(color, topleft)
            else: return BoxSprite(color, topleft)
        else:
            if f <= 0.006: return DiamondSprite("diamond", topleft)
            elif f <= 0.20: return BreakBoxSprite(color, topleft)
            else: return BoxSprite(color, topleft)

# Find all instances of a given type in a sequence.
def instances_in(sequence, typ):
    return list(filter(lambda x: isinstance(x, typ), sequence))

class SpriteBox(Sprite):
    def __init__(self, color, type = ""):
        Sprite.__init__(self)
        self._image = load.block(color, type)
        self._image.set_colorkey(self._image.get_at([0, 0]), RLEACCEL)
        self.image = transform.scale(self._image,
                                     [self.size[0] * 32, self.size[1] * 32])
        self._ckey = self._image.get_colorkey()
        self.rect = self.image.get_rect(topleft = [self.x * 32, self.y * 32])
        self._btime = 0

    def fall(self, field): self.rect.topleft = [self.x * 32, self.y * 32]
    def move(self, xy): self.rect.topleft = [xy[0] * 32, xy[1] * 32]

    def update(self, time):        
        if self.crashed and time > self._btime:
            for i in range(15):
                start = random.randrange(4)
                end = random.randrange(4)
                if end == start: end = (end + 1) % 4
                
                if start == 1: start = [random.randrange(32), 0]
                elif start == 2: start = [random.randrange(32), 31]
                elif start == 3: start = [0, random.randrange(32)]
                elif start == 0: start = [31, random.randrange(32)]
                
                if end == 1: end = [random.randrange(32), 0]
                elif end == 2: end = [random.randrange(32), 31]
                elif end == 3: end = [0, random.randrange(32)]
                elif end == 0: end = [31, random.randrange(32)]

                pygame.draw.line(self._image, self._ckey, start, end)

            self.image = transform.scale(self._image,
                                         [self.size[0] * 32,
                                          self.size[1] * 32])
            self._btime = time + 20


# A basic box type; grows when put near other boxes in the field.
class Box(AbstractBox):
    # Resize ourselve, and put ourself in the right place in the
    # field. Note that our topleft doesn't change, since the most
    # topleft gem is always the one resized.
    def set_size(self, size, field):
        self.size = size
        for x in range(self.x, self._get_right()):
            for y in range(self.y, self._get_bottom()):
                field[y][x] = self

    # Try to form larger blocks; return the blocks removed from merging.
    def try_merge(self, field):
        if self.size == [1,1]:
            # Special case the 1,1 size, because we don't want 2x1 or 1x2
            # blocks forming.
            if self.x == field.width - 1 or self.y == field.height - 1:
                return []
            b1 = field[self.y][self.x + 1]
            b2 = field[self.y + 1][self.x]
            b3 = field[self.y + 1][self.x + 1]
            if (isinstance(b1, Box) and isinstance(b2, Box) and
                isinstance(b3, Box) and
                self.color == b1.color == b2.color == b3.color and
                b1.size == b2.size == b3.size == [1, 1]):
                # Remove the sprites from the field and its sprite group
                b1.remove_from(field)
                b2.remove_from(field)
                b3.remove_from(field)
                self.set_size([2, 2], field)
                return [b1, b2, b3]
        else:
            # test all four directions
            blocks = self._adj_top(field)
            # there are blocks above us, and they are not special blocks,
            # and they are the same color, and they all share the same
            # top location, the leftmost border and rightmost border
            # match up with ours. We need these last checks to avoid
            # oddly shaped structures like
            # XXYY merging into XZZY
            #  ZZ                ZZ
            if (len(blocks) == self.size[0] and
                len(instances_in(blocks, Box)) == len(blocks) and
                [b.color for b in blocks].count(self.color) == len(blocks) and
                not [b for b in blocks if b.y != blocks[0].y] and
                blocks[0].x == self.x and
                blocks[-1]._get_right() == self._get_right()):
                dead = []
                for b in blocks:
                    b.remove_from(field)
                    dead.append(b)
                bot = self._get_bottom()
                self.y = blocks[0].y
                self.set_size([self.size[0], bot - self.y], field)
                return dead

            blocks = self._adj_bottom(field)
            if (len(blocks) == self.size[0] and
                len(instances_in(blocks, Box)) == len(blocks) and
                [b.color for b in blocks].count(self.color) == len(blocks) and
                not [b for b in blocks if
                     b._get_bottom() != blocks[0]._get_bottom()] and
                blocks[0].x == self.x and
                blocks[-1]._get_right() == self._get_right()):
                dead = []
                for b in blocks:
                    b.remove_from(field)
                    dead.append(b)
                bot = blocks[0]._get_bottom()
                self.set_size([self.size[0],  bot - self.y], field)
                return dead

            blocks = self._adj_left(field)
            if (len(blocks) == self.size[1] and
                len(instances_in(blocks, Box)) == len(blocks) and
                [b.color for b in blocks].count(self.color) == len(blocks) and
                not [b for b in blocks if b.x != blocks[1].x] and
                blocks[0].y == self.y and
                blocks[-1]._get_bottom() == self._get_bottom()):
                dead = []
                for b in blocks:
                    b.remove_from(field)
                    dead.append(b)
                right = self._get_right()
                self.x = blocks[0].x
                self.set_size([right - self.x, self.size[1]], field)
                return dead

            blocks = self._adj_right(field)
            if (len(blocks) == self.size[1] and
                len(instances_in(blocks, Box)) == len(blocks) and
                [b.color for b in blocks].count(self.color) == len(blocks) and
                not [b for b in blocks if
                     b._get_right() != blocks[0]._get_right()] and
                blocks[0].y == self.y and
                blocks[-1]._get_bottom() == self._get_bottom()):
                dead = []
                for b in blocks:
                    b.remove_from(field)
                    dead.append(b)
                right = blocks[0]._get_right()
                self.set_size([right - self.x, self.size[1]], field)
                return dead

        return []

# A version of the above that can be displayed on the screen.
class BoxSprite(Box, SpriteBox):
    def __init__(self, color, topleft):
        Box.__init__(self, color, topleft)
        SpriteBox.__init__(self, color, "")

    def move(self, xy):
        Box.move(self, xy)
        self.rect.topleft = [self.x * 32, self.y * 32]

    def fall(self, field):
        Box.fall(self, field)
        SpriteBox.fall(self, field)

    def set_size(self, size, field):
        Box.set_size(self, size, field)
        self.image = load.gem(self.color, self.size[0], self.size[1])
        self.rect = self.image.get_rect(topleft = [self.x * 32, self.y * 32])

# A gem that breaks gems of the same color when it lands by them.
class BreakBox(AbstractBox):
    # Called by the field to see if we can break anything.
    # THIS DOESN'T BREAK SELF! self will break when its
    # crashed method gets called by the gem *it* breaks.
    def try_crash(self, field):
        broken = 0
        for box in self.adjacent(field):
            broken += box.crash(field, self, True)
        return broken

class BreakBoxSprite(BreakBox, SpriteBox):
    def __init__(self, color, topleft):
        BreakBox.__init__(self, color, topleft)
        SpriteBox.__init__(self, color, "-crash")

    def fall(self, field):
        BreakBox.fall(self, field)
        SpriteBox.fall(self, field)

    def move(self, xy):
        BreakBox.move(self, xy)
        SpriteBox.move(self, xy)

# Now when you say special...
# These are the blocks used in "combat" mode, they don't form crystals,
# and when they break you get the 'item' in them.
class Special(AbstractBox):
    names = ["", "cleared", "reversed", "flipped", "blinking",
             "incoming", "scrambled"]
    
    def __init__(self, color, topleft, special = None):
        AbstractBox.__init__(self, color, topleft)
        self.special = (special or random.randint(1, 6))

    def crash(self, field, gem, immed = False):
        if not self.crashed:
            v = AbstractBox.crash(self, field, gem, immed)
            if self.crashed: field.pick_up(self.special)
        else: v = 0
        return v

class SpecialSprite(Special, SpriteBox):
    def load(cls, type):
        if type == SCRAMBLE:
            return textfx.shadow("? ?", 20, [255, 255, 255])
        else:
            fn = ["", "clear", "reverse", "flip", "blink", "gray"][type]
            return load.image("special-%s.png" % fn)

    load = classmethod(load)

    def __init__(self, color, topleft):
        Special.__init__(self, color, topleft)
        SpriteBox.__init__(self, color, "")
        self.image.blit(SpecialSprite.load(self.special), [6, 6])

    def fall(self, field):
        Special.fall(self, field)
        SpriteBox.fall(self, field)

    def move(self, xy):
        Special.move(self, xy)
        SpriteBox.move(self, xy)

# Diamonds break all of the color they land on, but "specially" -
# no bonuses for larger gems. So we just count the number of
# spaces of that color on the field, and then remove them.
# Extend BreakBox so that we find Diamonds when we filter for
# break boxes to crash.
class Diamond(BreakBox):
    def try_crash(self, field):
        broken = 0
        self.crashed = True
        if self.y == field.height - 1:
            field.tech_bonus()
            return 0
        else: color = field[self.y + 1][self.x].color

        for y in range(field.height):
            for x in range(field.width):
                box = field[y][x]
                if (isinstance(box, Box) or isinstance(box, BreakBox) or
                    isinstance(box, Special)):
                    if box.color == color:
                        broken += 1
                        box.crashed = True
                elif isinstance(box, TickBox):
                    if box.color == color:
                        broken += 0.5
                        box.crashed = True
        return broken

class DiamondSprite(Diamond, SpriteBox):
    def __init__(self, color, topleft):
        Diamond.__init__(self, color, topleft)
        SpriteBox.__init__(self, "diamond", "")

    def fall(self, field):
        Diamond.fall(self, field)
        SpriteBox.fall(self, field)

    def move(self, xy):
        Diamond.move(self, xy)
        SpriteBox.move(self, xy)

# Box with a counter on it. When it counts down, it replaces itself
# with a normal Box.
class TickBox(AbstractBox):
    def __init__(self, color, topleft):
        AbstractBox.__init__(self, color, topleft)
        self._time_left = 5

    def _get_time_left(self): return self._time_left
    def _set_time_left(self, timeleft):
        self._time_left = timeleft
        self._render()

    time_left = property(_get_time_left, _set_time_left)

    def tick(self):
        self.time_left -= 1

    def _render(self): pass

    def crash(self, field, gem, immed = False):
        # Tick boxes don't crash adjacent boxes, though they do
        # get destroyed themselves. They are also destroyed irrespective
        # of color.
        if not self.crashed and not immed:
            self.crashed = True
            return 0.5
        else: return 0.0

class TickBoxSprite(TickBox, SpriteBox):
    numerals = []

    def __init__(self, color, topleft):
        TickBox.__init__(self, color, topleft)
        SpriteBox.__init__(self, color, "")
        if not TickBoxSprite.numerals:
            f = pygame.font.Font(None, 32)
            for i in range(6):
                img = textfx.shadow(str(i + 1), f, [200, 200, 200])
                TickBoxSprite.numerals.append(img)
        self._render()

    def move(self, xy):
        TickBox.move(self, xy)
        SpriteBox.move(self, xy)

    def fall(self, field):
        TickBox.fall(self, field)
        SpriteBox.fall(self, field)

    def _render(self):
        self.image = pygame.Surface([32, 32])
        self.image.blit(self._image, [0, 0])
        t = TickBoxSprite.numerals[self._time_left - 1]
        self.image.blit(t, t.get_rect(center = [15, 15]))
        self.image.set_colorkey(self.image.get_at([0, 0]))        
        self.rect = self.image.get_rect(topleft = [self.x * 32, self.y * 32])



# game.py -- main game loop, player handling, field display, etc.
# Copyright 2004 Joe Wreschnig <piman@sacredchao.net>
# Released under the terms of the GNU GPL v2.
__revision__ = "$Id: game.py 317 2006-01-12 21:19:27Z piman $"

import os
import random
import pygame; import pygame_ext

from pygame.sprite import Sprite, RenderUpdates
from pygame.color import Color
from pygame import mixer, transform

from functools import cmp_to_key

from constants import *

import events; from events import Event, EventManager
from boxes import Box, TickBox, BreakBox, Special, Diamond, BoxGen
from boxes import BoxSprite, TickBoxSprite, SpecialSprite

from characters import Character

import textfx
import config
import wipes
import load
import util

# Manage boning counter for single player mode. Keep track of the number
# of turns and blocks left, and drop the blocks (randomly) when that
# number reaches 0. This replaces DropPattern in single player.
class TensionBoner(object):
    def __init__(self):
        self.turns_left = 10
        self.turns_left_max = 10
        self.next_boning = 2
        self.next_boning_max = 2

    # A piece hit the bottom
    def tick(self):
        self.turns_left -= 1

    # The player cleared some gems.
    def debone(self, player, val):
        self.next_boning -= val
        if self.next_boning <= 0:
            player.score.score += (-50 * self.next_boning)
            self.next_boning_max = self.next_boning_max * (5.0/3.0)
            self.turns_left_max += 2
            self.next_boning = int(self.next_boning_max)
            player.score.score += 100 * self.turns_left
            self.turns_left = self.turns_left_max
        return 0

    # No turns left. Drop random blocks.
    # This shares a lot of code with the DropPattern algorithm.
    def bone(self):
        count = self.next_boning
        rows = []
        while count >= 15:
            rows.append([random.choice(COLORS) for i in range(15)])
            count -= 15
        if count != 0:
            possible = list(range(15))
            to_fill = []
            rows.append([random.choice(COLORS) for i in range(15)])
            while count > 0 and len(possible) > 0:
                to_fill.append(random.choice(possible))
                possible.remove(to_fill[-1])
                count -= 1

            for i in range(len(rows[-1])):
                if i not in to_fill: rows[-1][i] = None
            
        self.next_boning_max = self.next_boning_max * (1.5)
        self.turns_left_max += 2
        self.next_boning = int(self.next_boning_max)
        self.turns_left = self.turns_left_max
        return rows

# Display messages over the field when things happen. Usually a chain,
# but sometimes other things.
class ChainText(Sprite):
    def __init__(self, center):
        Sprite.__init__(self)
        self._center = center
        self._font = pygame.font.Font(None, 48)
        self._blank = pygame.Surface([0, 0])
        self._ntime = 0
        self.image = self._blank
        self.rect = self.image.get_rect()

    def set_chain(self, chain):
        if chain > 1: self.set_text("%d chain!" % chain)

    def set_text(self, text):
        self._ntime = pygame.time.get_ticks() + 1000
        self.image = self._font.render(text, True, [255, 255, 255])
        self.rect = self.image.get_rect()
        self.rect.center = self._center

    chain = property(None, set_chain)
    text = property(None, set_text)

    def update(self, time):
        if self._ntime and time > self._ntime:
            self.image = self._blank
            self.rect = self.image.get_rect()
            self._ntime = None

# The boxes that fall down with time.
class FallingBoxes(Sprite):
    # Orientation/rotation rules:
    # There are 4 possible orientations for the boxes:
    #  2
    #  1 21  1  12    (left to right, rotate counter-clockwise)
    #        2
    #
    #  0  1  2  3

    def __init__(self, box1, box2, field):
        Sprite.__init__(self)
        self.y = 0
        self.x = int(field.width / 2)
        self._box1 = box1
        self._box2 = box2
        self._orient = 0
        self.locked = False
        self._field = field
        self._ntime = None
        self.bonus = False
        self._render()

    # This sprite just appeared on the screen and is now falling.
    def start(self, time): self._ntime = time + 750

    # Predicates to see if we're blocked in the directions we want to go.
    def _is_blocked_left(self):
        l = self._get_left()
        return ((l == 0) or
                (self._field[self._get_top()][l - 1] or
                 self._field[self._get_bottom()][l - 1]))

    def _is_blocked_right(self):
        r = self._get_right()
        return ((r == self._field.width - 1) or
                (self._field[self._get_top()][r + 1] or
                 self._field[self._get_bottom()][r + 1]))

    def _is_blocked_down(self):
        b = self._get_bottom()
        return ((b == self._field.height - 1) or
                (self._field[b + 1][self._get_left()] or
                 self._field[b + 1][self._get_right()]))

    # Actually go right or left, after checking.
    def go_left(self):
        if not (self.locked or self._is_blocked_left()): self.x -= 1
    def go_right(self):
        if not (self.locked or self._is_blocked_right()): self.x += 1

    # Go down. If we can't go down, lock ourself in place.
    def go_down(self):
        if not self._is_blocked_down():
            self._ntime = pygame.time.get_ticks() + self._field.speed
            self.y += 1
        else:
            self.locked = True
            self.bonus = True

    def _get_top(self): return max(0, self.y - int(self._orient == 0))
    def _get_left(self): return self.x - int(self._orient == 1)
    def _get_bottom(self): return self.y + int(self._orient == 2)
    def _get_right(self): return self.x + int(self._orient == 3)

    # Drop our contained boxes onto the field and let them fall.
    def deposit(self):
        if self._orient == 0:
            self._field.add_box([self.x, self.y], self._box1)
            if self.y != 0:
                self._field.add_box([self.x, self.y - 1], self._box2)
        elif self._orient == 1:
            self._field.add_box([self.x, self.y], self._box1)
            self._field.add_box([self.x - 1, self.y], self._box2)
        elif self._orient == 2:
            self._field.add_box([self.x, self.y], self._box1)
            self._field.add_box([self.x, self.y + 1], self._box2)
        elif self._orient == 3:
            self._field.add_box([self.x, self.y], self._box1)
            self._field.add_box([self.x + 1, self.y], self._box2)
        return self._box1, self._box2

    # When a falling block can't rotate properly, the boxes are
    # swapped instead of rotated.
    def _swap(self): self._box1, self._box2 = self._box2, self._box1

    # Try to rotate the piece; push it away from a wall if we need to;
    # swap if we can't. Takes the new orientation and as an argument.
    def _rotate(self, new_o):
        if self.locked: return
        blocked = False
        pushable = False
        dx, dy = 0, 0
        if new_o == 3:
            blocked = (self._get_right() == self._field.width - 1 or
                       self._field[self.y][self.x + 1] is not None)
            pushable = (blocked and
                        self._get_left() != 0 and
                        self._field[self.y][self.x - 1] is None)
            dx = -1
        elif new_o == 0:
            blocked = (self._field[self.y - 1][self.x] is not None)
            pushable = (blocked and
                        self._field[self.y + 1][self.x] is None)
            dy = 1
        elif new_o == 1:
            blocked = (self._get_left() == 0 or
                       self._field[self.y][self.x - 1] is not None)
            pushable = (blocked and
                        self._get_right() != self._field.width - 1 and
                        self._field[self.y][self.x + 1] is None)
            dx = 1
        elif new_o == 2:
            blocked = (self.y == self._field.height - 1 or
                       self._field[self.y + 1][self.x] is not None)

        FallingBoxes.rotate_sound.play()
        if blocked and not pushable:
            self._swap()
        else:
            if pushable:
                self.x += dx
                self.y += dy
            self._orient = new_o
        self._render()

    # Rotate clockwise or counterclockwise.
    def rot_cw(self): self._rotate((self._orient - 1) % 4)
    def rot_cc(self): self._rotate((self._orient + 1) % 4)

    def update(self, time):
        if self._ntime is None: return
        if time > self._ntime:
            if self._is_blocked_down():
                self.locked = True
            else:
                self.y += 1
                self._ntime = time + self._field.speed

        self.rect.left = self.x * 32
        if self._orient == 1: self.rect.left -= 32
        self.rect.top = self.y * 32
        if self._orient == 0: self.rect.top -= 32
        off = float(self._ntime - time) / self._field.speed
        self.rect.top -= int(off * 32)

    # Draw the boxes.
    def _render(self):
        if self._orient & 1 == 0: self.image = pygame.Surface([32, 64])
        else: self.image = pygame.Surface([64, 32])
        self.image.set_colorkey(self.image.get_at([0, 0]))

        if self._orient == 0:
            self.image.blit(self._box1.image, [0, 32])
            self.image.blit(self._box2.image, [0, 0])
        elif self._orient == 1:
            self.image.blit(self._box1.image, [32, 0])
            self.image.blit(self._box2.image, [0, 0])
        elif self._orient == 2:
            self.image.blit(self._box1.image, [0, 0])
            self.image.blit(self._box2.image, [0, 32])
        elif self._orient == 3:
            self.image.blit(self._box1.image, [0, 0])
            self.image.blit(self._box2.image, [32, 0])

        self.rect = self.image.get_rect()

# A simple non-graphical playing field.
class BasicField(object):
    def __init__(self, width, height):
        self.width = width
        self.height = height
        self.max_gemsize = 0
        self.max_chain = 0
        self._field = [[None] * width for i in range(height)]

    def __getitem__(self, i): return self._field[i]
    def __iter__(self): return iter(self._field)

    def add_box(self, xy, box):
        box.move(xy)
        self._field[box.y][box.x] = box

    # Move all pieces down one step; return True if pieces moved
    # or False otherwise.
    def fall(self):
        could_move = False
        for i in range(len(self._field) - 2, -1, -1):
            # Things on the first row won't fall, so we don't
            # need to check it.
            for j, box in enumerate(self._field[i]):
                if (box and not box.is_blocked_down(self)):
                    box.fall(self)
                    could_move = True
        return could_move

    # Tick all ticking boxes; replace them if their time runs out.
    def tick(self, Box = Box):
        tickers = list(filter(lambda b: isinstance(b, TickBox),
                         util.flatten(self._field)))
        extickers = []
        for t in tickers:
            t.tick()
            if t.time_left == 0:
                t.remove_from(self)
                extickers.append(t)
                b = Box(t.color, [t.x, t.y])
                self.add_box([t.x, t.y], b)
        return tickers, extickers

    def pick_up(self, special): pass

    # BasicField doesn't know what a "player" is, so it can't
    # give a bonus.
    def tech_bonus(self): pass

    def cmp(self, x, y):
        return (x > y) - (x < y)

    # Try to merge boxes into bigger ones; return True if one
    # merged successfully (only merges one)
    def merge(self):
        boxes = list(filter(lambda b: isinstance(b, Box),
                       util.flatten(self._field)))
        boxes.sort(key=cmp_to_key(lambda a, b: self.cmp(b.y, a.y) and self.cmp(a.x, b.x)))
        for box in boxes:
            dead = box.try_merge(self)
            if len(dead) > 0: return dead
        return []

    # Break everything that will break; return the "value" of
    # the broken 
    def breaking(self):
        breakers = list(filter(lambda b: isinstance(b, BreakBox),
                          util.flatten(self._field)))
        killed_count = sum([b.try_crash(self) for b in breakers
                            if not b.crashed])
        dead = list(filter(lambda b: b and b.crashed, util.flatten(self._field)))
        for box in dead: box.remove_from(self)
        return killed_count, dead

# Graphics for the above field, and a state machine to manage gameplay.
class Field(BasicField, Sprite):
    # The six stages of each "turn":
    # 1. A piece is dropping down and the player can control it (FALLING).
    # 2. The piece hits. Any loose pieces must be dropped downwards.
    #    (FIXING1)
    # 3. TICKING. Any timer block, must tick.
    # 4. MERGING. Large gems form out of smaller ones.
    # 5. BREAKING. Any breaking gems take effect.
    # 6. Stuff breaks. Loose pieces drop (FIXING2).
    # 7. Go to 4) until 6 makes no changes.
    # 8. Drop blocks from opponent's clears. (BONING)
    # 9. Move blocks down one if possible. (BONING2)    
    #    Repeat until player is boned again or no movement.
    # 11. Insert the new falling block. Start again from 1).
    FALLING, FIXING1, TICKING, MERGING, BREAKING, FIXING2, BONING, BONING2 = range(8)

    def __init__(self, player, width = 6, height = 13):
        BasicField.__init__(self, width, height)
        Sprite.__init__(self)
        self._sprites = RenderUpdates()
        self.falling = None
        self._break_time = None
        self.speed = config.getint("settings", "speed")
        self._field = [[None] * width for i in range(height)]
        self._player = player
        self.state = Field.FIXING1
        self._insrow = int((width / 2))
        self._ntime = pygame.time.get_ticks()
        self._broke_any_last = False
        self._killed = 0
        self._disease = HEALTHY
        self._disease_end = 0
        self._chain = 0

        self.rect = Rect([player.topleft, [width*32 + 20, height*32 + 20]])
        # Darken a bit of the screen and use that for our background
        black = pygame.Surface([width * 32, height * 32]).convert()
        black.set_alpha(160)
        screen = pygame.display.get_surface()
        bg = pygame.Surface([width * 32, height * 32]).convert()
        bg.blit(screen, [-player.topleft[0] - 10, -player.topleft[1] - 10])
        bg.blit(black, [0, 0])

        self._outline = player.char.border(bg)
        self._chain_spr = ChainText([int((width / 2.0) * 32) + 10,
                                     int((height / 2.0) * 32) + 10])

        self._render()
        self._state_calls = [self._state_falling,
                             self._state_fixing1,
                             self._state_ticking,
                             self._state_merging,
                             self._state_breaking,
                             self._state_fixing2,
                             self._state_boning,
                             self._state_boning2
                             ]

    # Someone used a special on us...
    def inflict(self, disease):
        if disease == SCRAMBLE:
            self._player._next_disp.scramble()
        elif disease == GRAY:
            row = ["gray", "gray", None, None, None, None]
            random.shuffle(row)
            self._player.enqueue([row])
        elif disease == BLINK:
            self._disease = disease
            self._disease_end = pygame.time.get_ticks() + 30000
        else:
            self._disease = disease
            self._disease_end = pygame.time.get_ticks() + 10000
        self._chain_spr.text = Special.names[disease] + "!"

    def _get_image(self):
        if self._disease == FLIP:
            return transform.flip(self._image, False, True)
        elif self._disease == BLINK:
            # V*V*(3-V-V), V is a value 0-1, flattens the curve around 0-1
            # without using sin or cos.
            t = (pygame.time.get_ticks() % 3000) / 1500.0
            t = abs(t - 1)
            t = t * t * (3 - t - t)
            self._image.set_alpha(int(256 * t))
            return self._image
        else: return self._image

    image = property(_get_image)

    def clear(self):
        self._field = [[None] * self.width for i in range(self.height)]
        self._sprites.empty()

    def _render(self):
        image = pygame.Surface([self.width * 32, self.height * 32])
        image.set_colorkey(image.get_at([0, 0]))
        self._sprites.draw(image)
        if self.falling:
            image.blit(self.falling.image, self.falling.rect)
        self._image = pygame.Surface([self.width * 32 + 20,
                                      self.height * 32 + 20])
        self._image.blit(self._outline, [0, 0])
        self._image.blit(image, [10, 10])
        self._image.blit(self._chain_spr.image, self._chain_spr.rect)

    def pick_up(self, special):
        self._player.pick_up(special)

    # Remove has to be done by the box itself, but this we can do here...
    def add_box(self, xy, box):
        BasicField.add_box(self, xy, box)
        self._sprites.add(box)

    def move(self, ev):
        if self._disease == REVERSE:
            ev = { LEFT: RIGHT, RIGHT: LEFT, UP: DOWN, DOWN: UP,
                   ROT_CW: ROT_CC, ROT_CC: ROT_CW }.get(ev, ev)

        if self.state == Field.FALLING and self.falling:
            if ev == LEFT: self.falling.go_left()
            elif ev == RIGHT: self.falling.go_right()
            elif ev == DOWN: self.falling.go_down()
            elif ev == ROT_CC: self.falling.rot_cc()
            elif (ev == ROT_CW or
                  (ev == UP and
                   config.getboolean("settings", "rotate_on_up"))):
                  self.falling.rot_cw()

    def _state_falling(self, time):
        if not self.falling:
            if self._field[0][self._insrow] is not None: self._player.die()
            self.falling = self._player.get_next_falling(time)
            self._player.tick()
        self.falling.update(time)
        if self.falling.locked:
            if self.falling.bonus:
                self._player.score.score += 10
            self.speed = max(250, self.speed - 5)
            self.state = Field.FIXING1
            self._ntime = time + 20
            self.falling.deposit()
            self.falling = None

    def _state_fixing1(self, time):
        # Strictly speaking, this loop is overkill, since we know
        # that at most one thing will fall...
        # So, if we need to, optimize this by using the return
        # value of deposit later.
        if time > self._ntime: 
            self._ntime = time + 20
            if not self.fall(): self.state = Field.TICKING

    def _state_ticking(self, time):
        tickers, extickers = self.tick(BoxSprite)
        if len(tickers) != 0: TickBox.sound.play()
        for t in extickers: t.kill()
        self.state = Field.MERGING

    def _state_merging(self, time):
        dead = self.merge()
        for spr in dead: spr.kill()
        if len(dead) == 0: self.state = Field.BREAKING

    def _state_breaking(self, time):
        if self._break_time is None:
            killed_count, dead = self.breaking()
            self._player.score.score +=  killed_count * 100
            self._killed += killed_count * (1 + self._chain / 2.0)
            self._broke_any_last = (len(dead) != 0)
            self._dead = dead
            if len(dead) > 0:
                BreakBox.sound.play()
                self._break_time = time + 100
            else:
                self.state = Field.FIXING2
        elif time > self._break_time:
            for spr in self._dead: spr.kill()
            self._break_time = None
            self.state = Field.FIXING2

    def _state_fixing2(self, time):
        if not self._broke_any_last:
            # Process pending boning drops for other players, and then
            # switch to our own boning state.
            val = int(self._killed)
            self._killed = 0
            self._player.distribute_boning(val)
            self.state = Field.BONING

        # *This* loop is not overkill, since stuff will keep breaking
        # and falling...
        elif time > self._ntime: 
            self._ntime = time + 20
            if not self.fall():
                self._chain += 1
                if self._chain > 1:
                    self.max_chain = max(self.max_chain, self._chain)
                self._chain_spr.chain = self._chain
                self.state = Field.MERGING

    def _state_boning(self, time):
        for x, box in enumerate(self._player.get_boning_row()):
            if box:
                if self._field[0][x] is None:
                    self._field[0][x] = box
                    box.move([x, 0])
                    self._sprites.add(box)
        self.state = Field.BONING2
        self._ntime = time

    def tech_bonus(self):
        self._chain_spr.text = "tech bonus!"
        self._player.score.score += 1000

    def _state_boning2(self, time):
        if time > self._ntime: 
            self._ntime = time + 50
            could_move = self.fall()

            if self._player.is_boned():
                self.state = Field.BONING
            elif not could_move:
                self._chain = 0
                self._chain_spr.chain = self._chain
                self.state = Field.FALLING

    def update(self, time):
        if self._disease != HEALTHY and time > self._disease_end:
            if self._disease == BLINK: self.image.set_alpha(256)
            self._disease = HEALTHY
        self._chain_spr.update(time)
        self._sprites.update(time)
        self._state_calls[self.state](time)
        self._render()

# Display the piece that will be dropped next.
class NextBlock(Sprite):
    def __init__(self, topleft, char):
        Sprite.__init__(self)
        self._base = char.border([42, 110])
        nxt = textfx.shadow("next:", 20)
        self._base.blit(nxt, nxt.get_rect(center = [32, 20]))
        self._end_scramble = 0
        
        self.image = self._base
        self.rect = self.image.get_rect()
        self.rect.topleft = topleft

    def scramble(self): self._end_scramble = 10

    def set_next(self, block):
        self.image = self._base.convert()
        if self._end_scramble > 0:
            block = pygame.Surface([32, 64])
            block.blit(load.block(random.choice(COLORS)), [0, 0])
            block.blit(load.block(random.choice(COLORS)), [0, 32])
            block.set_colorkey(block.get_at([0, 0]))
            self._end_scramble -= 1
        else:
            block = block.image
        self.image.blit(block, [15, 40])

# Display a number in a box with a short title.
class CountSprite(Sprite):
    def __init__(self, topleft, text, char):
        Sprite.__init__(self)
        self._base = char.border([40, 40])
        inc = textfx.shadow(text, 20)
        self._count = 0
        self._base.blit(inc, inc.get_rect(center = [30, 20]))
        self.image = self._base
        self.rect = self.image.get_rect()
        self.rect.topleft = topleft
        self._set_count(0)

    def _get_count(self): return self._count
    def _set_count(self, count):
        self._count = max(0, count)
        self.image = self._base.convert()
        txt = textfx.shadow(str(self._count), 30)
        self.image.blit(txt, txt.get_rect(center = [30, 40]))

    count = property(_get_count, _set_count)

    def reset(self): self._set_count(0)

# Hold the "special" attack graphic a player currently has.
class SpecialHolder(Sprite):
    def __init__(self, topleft, char):
        Sprite.__init__(self)
        self._base = char.border([40, 40])
        text = textfx.shadow("att.:", 20)
        self._special = HEALTHY
        self._base.blit(text, text.get_rect(center = [30, 20]))
        self.image = self._base
        self.rect = self.image.get_rect()
        self.rect.topleft = topleft

    def _get_special(self): return self._special
    def _set_special(self, special):
        if special:
            self.image = self._base.convert()
            self._special = special
            self.image.blit(SpecialSprite.load(self.special), [20, 30])
        else:
            self.image = self._base

    special = property(_get_special, _set_special)

# Display the player's score.
class Score(Sprite):
    def __init__(self, topleft, char):
        Sprite.__init__(self)
        self._base = char.border([120, 40])
        title = textfx.shadow("score:", 20)
        self._score = 0
        self._base.blit(title, title.get_rect(topleft = [15, 10]))
        self.image = self._base
        self.rect = self.image.get_rect()
        self.rect.topleft = topleft
        self.score = 0

    def _get_score(self): return self._score
    def _set_score(self, score):
        self._score = score
        self.image = self._base.convert()
        txt = textfx.shadow("%09d" % self._score, 30)
        self.image.blit(txt, txt.get_rect(midright = [125, 38]))

    score = property(_get_score, _set_score)

class AbstractPlayer(object):
    def __init__(self, pid, char):
        self._pid = pid
        self.char = char
        self.is_ai = False
        self.centerx = self.topleft[0] + 96 + 10
        self.score = Score([self.topleft[0] + 72, self.topleft[1] + 520],
                            self.char)

    def win(self):
        self.score.score += 500 * self.field.max_chain
        self.score.score += 100 * self.field.max_gemsize

    def win_data(self):
        statbox = self.char.border([160, 230])
        oldscore = self.score.score - (500 * self.field.max_chain +
                                       100 * self.field.max_gemsize)
        t1l = textfx.shadow("Score:", 30)
        t1r = textfx.shadow("%d" % oldscore, 26)
        t2l = textfx.shadow("Largest Crystal:", 30)
        t2r = textfx.shadow("%d x 100 = +%d" % (self.field.max_gemsize,
                                          self.field.max_gemsize * 100), 26)
        t3l = textfx.shadow("Longest Chain:", 30)
        t3r = textfx.shadow("%d x 500 = +%d" % (self.field.max_chain,
                                         500 * self.field.max_chain), 26)
        t4l = textfx.shadow("Total Score:", 30)
        t4r = textfx.shadow("%d" % self.score.score, 30)
        statbox.blit(t1l, [15, 15])
        statbox.blit(t2l, [15, 65])
        statbox.blit(t3l, [15, 120])
        statbox.blit(t4l, [15, 185])
        statbox.blit(t1r, [statbox.get_width() - 20 - t1r.get_width(), 40])
        statbox.blit(t2r, [statbox.get_width() - 20 - t2r.get_width(), 90])
        statbox.blit(t3r, [statbox.get_width() - 20 - t3r.get_width(), 150])
        statbox.blit(t4r, [statbox.get_width() - 20 - t4r.get_width(), 210])
        return statbox

    def distribute_boning(self, count): pass

    def get_boning_row(self):
        if len(self._pending_boning) > 0:
            ret = self._pending_boning.pop(0)
            self.drop_count.count -= (len(ret) - ret.count(None))
            return ret
        else: return []

    def die(self):
        print("Player %d is dead!" % (self._pid + 1))
        self.dead = True

    def tick(self): pass

    # Called when a game is started. Players can persist across games.
    def start(self, seed, others):
        self._rand = random.Random(seed)
        self._pending_boning = []
        self.dead = False
        self._gen = BoxGen(self._rand, self.colors,
                                 isinstance(self, SinglePlayer))
        self.others = list(filter(lambda x: x is not self, others))

        self.dead_image = pygame.Surface([self.field.width * 32,
                                          self.field.height * 32])
        img = load.block("gray")
        for i in range(0, self.dead_image.get_width(), 32):
            for j in range(0, self.dead_image.get_height(), 32):
                self.dead_image.blit(img, [i, j])
        self.dead_image = self.char.border(self.dead_image)

        self._next = FallingBoxes(self._gen.get([-1, -1]),
                                  self._gen.get([-1, -1]),
                                  self.field)
        self._next.rect = self._next.image.get_rect()
        self._next.rect.topleft = [self.topleft[0] + 30, 500]
        self._next_disp = NextBlock([self.topleft[0], self.topleft[1] + 450],
                                    self.char)
        self.sprites = RenderUpdates([self.field,self._next_disp,self.score])
        self._next_disp.set_next(self._next)

    def reset(self):
        self.dead = False

    def get_next_falling(self, time):
        n = self._next
        self._next = FallingBoxes(self._gen.get([-1, -1]),
                                  self._gen.get([-1, -1]),
                                  self.field)
        self._next.rect = self._next.image.get_rect()
        self._next.rect.topleft = [self.topleft[0] + 30, 500]
        self._next_disp.set_next(self._next)
        n.start(time)
        return n

    def is_boned(self): return (len(self._pending_boning) != 0)

    def use_special(self): pass

    def deboning(self, value): abstract
    def enqueue(self, value): abstract

    def update(self, time):
        self.sprites.update(time)

class VersusPlayer(AbstractPlayer):
    def __init__(self, pid, char):
        if pid == 0: self.topleft = [10, 0]
        elif pid == 1: self.topleft = [800 - 32 * 6 - 30, 0]
        self.colors = COLORS[:4]
        AbstractPlayer.__init__(self, pid, char)

    def start(self, seed, others):
        self.field = Field(self)
        self.special = None
        AbstractPlayer.start(self, seed, others)
        self.drop_count = CountSprite([self.topleft[0] + 73,
                                        self.topleft[1] + 450],
                                       "new:", self.char)
        self.sprites.add(self.drop_count)
        if config.getboolean("settings", "combat"):
            self._special_spr = SpecialHolder([self.topleft[0] + 152,
                                               self.topleft[1] + 450],
                                              self.char)
            self.sprites.add(self._special_spr)

    def enqueue(self, rows):
        for row in rows:
            self.drop_count.count += (len(row) - row.count(None))
            # Hacky -- the and acts like a conditional.
            row = [(c and TickBoxSprite(c, [-1, -1])) for c in row]
            self._pending_boning.append(row)

    def distribute_boning(self, value):
        value = (float(self.deboning(value)) / len(self.others))
        if value == 0: return
        for p in self.others:
            v = self.char.drop.multiply(value)
            p.enqueue(v)

    def pick_up(self, special):
        Special.sound.play()
        self._special_spr.special = self.special = special

    def use_special(self):
        if self.special is None: return
        elif self.special == CLEAR:
            self.field.clear()
            self.field._chain_spr.text = Special.names[CLEAR] + "!"
        elif self.special is not None:
            for p in self.others: p.field.inflict(self.special)
        self._special_spr.special = self.special = None
        

    def deboning(self, value):
        if len(self._pending_boning) == 0: return value

        rows = self._pending_boning
        total = sum([len(row) - row.count(None) for row in rows])
        half_time = False
        if value >= total: half_time = True
        while value > 0:
            if len(rows) == 0: break
            for i in range(len(rows[-1])):
                if rows[-1] is not None and value > 0:
                    rows[-1][i] = None
                    value -= 2
                    self.drop_count.count -= 1
            if rows[-1].count(None) == len(rows[-1]): rows.pop()
        if half_time:
            for row in rows:
                for box in row:
                    if isinstance(box, TickBox): box.time_left = 3
        return max(0, value)

class AIPlayer(VersusPlayer):
    AIS = { STUPID: "StupidAI",
            VEASY: "VeryEasyAI",
            EASY: "EasyAI",
            NORMAL: "NormalAI",
            HARD: "HardAI",
            INSANE: "InsaneAI" }
    
    def __init__(self, pid, char, ai_name = None):
        # I, for one, welcome our new robot overlords.
        VersusPlayer.__init__(self, pid, char)
        self.is_ai = True
        import ai
        if ai_name == None:
            ai_name = AIPlayer.AIS[config.getint("settings", "ai")]
        self._ai = ai.__dict__[ai_name](self)

    def start(self, *args):
        VersusPlayer.start(self, *args)
        self._iter = None
        self._goal = None
        self._moved = False
        self._aimove_ntime = 0

    def tick(self):
        self._ai.incoming(self.field, self.field.falling)
        self._iter = iter(self._ai)
        self._goal = None
        self._moved = False

    def update(self, time):
        VersusPlayer.update(self, time)
        if self._iter:
            move = next(self._iter)
            if move != None:
                self._iter = None
                self._goal = move
                self._aimove_ntime = time - 1

        if self._goal and time > self._aimove_ntime and self.field.falling:
            if self._goal[0] != self.field.falling._orient:
                events.post(Event(PLAYER, key = ROT_CC, player = self._pid))
            elif self._goal[1] < self.field.falling.x:
                events.post(Event(PLAYER, key = LEFT, player = self._pid))
            elif self._goal[1] > self.field.falling.x:
                events.post(Event(PLAYER, key = RIGHT, player = self._pid))
            else:
                self._goal = None
                self._moved = True
            self._aimove_ntime = time + self._ai.delta

        if (self._ai.drops and
            self._moved and time > self._aimove_ntime and self.field.falling):
            events.post(Event(PLAYER, key = DOWN, player = self._pid))
            self._aimove_ntime = time + self._ai.drops

class SinglePlayer(AbstractPlayer):
    def __init__(self, pid, char):
        self.topleft = [155, 0]
        self.colors = COLORS
        AbstractPlayer.__init__(self, pid, char)

    def tick(self):
        self.tension.tick()
        if self.tension.turns_left <= 0:
            self.enqueue(self.tension.bone())
        self.ttime.count = self.tension.turns_left
        self.drop_count.count = self.tension.next_boning

    def start(self, seed, others):
        self.field = Field(self, 15, 13)
        AbstractPlayer.start(self, seed, others)
        self.tension = TensionBoner()
        self.drop_count = CountSprite([self.topleft[0] + 73,
                                  self.topleft[1] + 450],
                                 "new:", self.char)
        self.ttime = CountSprite([self.topleft[0] + 150,
                                  self.topleft[1] + 450],
                                 "turns:", self.char)
        self.ttime.count = 10
        self.drop_count.count = 2
        self.sprites.add([self.drop_count, self.ttime])

    def distribute_boning(self, value):
        self.deboning(value)

    def enqueue(self, rows):
        for row in rows:
            # Hacky -- the and acts like a conditional.
            row = [(c and TickBoxSprite(c, [-1, -1])) for c in row]
            self._pending_boning.append(row)

    def deboning(self, value):
        self.tension.debone(self, value)
        self.drop_count.count = self.tension.next_boning
        self.ttime.count = self.tension.turns_left
        return 0

class FightSprite(Sprite):
    sounds = [load.sound("begin.wav")]
    for i in range(1, 10):
        sounds.append(load.sound("round%d.wav" % i))
    for snd in sounds: snd.set_volume(0.6)

    def __init__(self, match, matches):
        Sprite.__init__(self)
        self._start = pygame.time.get_ticks()
        self._played_round = False
        self._played_begin = False
        self._match = match + 1
        text = "Round %d/%d" % (self._match, matches)
        self._later = textfx.shadow("Begin!", 300, color = [255, 255, 255])
        self._later_rect = self._later.get_rect(center = [400, 300])
        self.image = textfx.shadow(text, 150, color = [255, 255, 255])
        self.rect = self.image.get_rect()
        self.rect.right = 0
        self.rect.centery = 300
        self._w = 800 + self.image.get_width()

    def update(self, time):
        if time - self._start < 2000:
            if not self._played_round:
                FightSprite.sounds[self._match].play()
                self._played_round = True
            self.rect.right = int(((time - self._start) / 2000.0) * self._w)
        elif time - self._start < 3000:
            if not self._played_begin:
                FightSprite.sounds[0].play()
                self._played_begin = True
            self.image = self._later
            self.rect = self._later_rect
        else:
            self.kill()

class Game(object):
    def __init__(self):
        self._sprites = RenderUpdates()
        self._screen = pygame.display.get_surface()
        self._em = EventManager()
        self._pickaxe = load.image("pickaxe.png")
        self._light_axe = self._pickaxe.copy()
        alphs = pygame.surfarray.pixels_alpha(self._light_axe)
        for row in alphs:
            for i in range(len(row)): row[i] = row[i] // 2
        del(alphs)

    def play(self, matches, wins):
        seed = pygame.time.get_ticks()
        screen = pygame.display.get_surface()

        for i in range(wins[0]):
            screen.blit(self._pickaxe, [225 + 70 * i, 400])
        for i in range(wins[0], (matches + 1) // 2):
            screen.blit(self._light_axe, [225 + 70 * i, 400])
        for i in range(wins[1]):
            screen.blit(self._pickaxe, [505 - 70 * i, 500])
        for i in range(wins[1], (matches + 1) // 2):
            screen.blit(self._light_axe, [505 - 70 * i, 500])
        pygame.display.update()

        for p in self._players: p.start(seed, self._players)
        em = EventManager()
        quit = False
        paused = False
        dim = pygame.Surface(screen.get_size(), 16)
        dim.set_alpha(128)
        f = pygame.font.Font(None, 48)
        txt = f.render("- paused -", True, [255, 255, 255])
        dim.blit(txt, txt.get_rect(center = [400, 300]))
        start_pause = 0
        prev_screen = None
        events.DELAY = 100
        pygame.key.set_repeat(100, 75)
        while (not quit and
               ((len(self._players) > 1 and
                 [p.dead for p in self._players].count(False) > 1)
                or (len(self._players) == 1 and not self._players[0].dead))):
            for ev in em.get():
                if ev.type == PLAYER:
                    if ev.player < len(self._players):
                        if ev.key == CONFIRM:
                            if pygame.time.paused:
                                pygame.event.post(Event(PAUSE))
                            else:
                                self._players[ev.player].use_special()
                        else:
                            self._players[ev.player].field.move(ev.key)
                elif ev.type == PAUSE:
                    pygame.time.toggle_pause()
                    if pygame.time.paused:
                        prev_screen = screen.convert()
                        screen.blit(dim, [0, 0])
                    else: screen.blit(prev_screen, [0, 0])
                    pygame.display.update()
                elif ev.type == QUIT:
                    quit = True
                    break
            if quit: break

            if not pygame.time.paused:
                for p in self._players:
                    p.update(pygame.time.get_ticks())

                for p in self._players:
                    pygame.display.update(p.sprites.draw(screen))

            pygame.time.clock.tick(60)
        pygame.key.set_repeat(200, 75)
        events.DELAY = 200

    def block_out(self):
        try: pygame.mixer.music.fadeout(1000)
        except: pass
        dead_snd = load.sound("death.wav")
        indices = [i for i, p in enumerate(self._players) if p.dead]
        dead_snd.play()
        t = start = pygame.time.get_ticks()
        while t - start < 1000:
            h = 400 - (400 * ((t - start) / 1000.0))
            self._screen.set_clip([[0, h], [800, 450 - h]])
            for i in indices:
                pygame.display.update(self._screen.blit(
                    self._players[i].dead_image,
                    self._players[i].topleft))
            pygame.time.clock.tick(60)
            t = pygame.time.get_ticks()
        self._screen.set_clip()

    def wait_reset(self, text, char):
        cont = True
        last = pygame.time.get_ticks()
        screen = pygame.display.get_surface()
        #pygame.mixer.music.load(os.path.join("music", "between.ogg"))
        #pygame.mixer.music.set_volume(0.5)
        #pygame.mixer.music.play(-1)
        while cont:
            t = pygame.time.get_ticks()
            for ev in self._em.get():
                if (ev.type == QUIT or
                    (ev.type == PLAYER and
                     (ev.key in [CONFIRM, ROT_CC, ROT_CW]))):
                    cont = False
            if t - last > 80:
                img = text.animate()
                border = char.border(img.get_size())
                border.blit(img, [10, 10])
                r = border.get_rect(center = [400, 100])
                pygame.display.update(screen.blit(border, r))
                last = t
            pygame.time.clock.tick(60)

        for p in self._players: p.reset()
        #pygame.mixer.music.stop()
        #pygame.mixer.music.set_volume(1.0)

class SingleGame(Game):
    def __init__(self):
        Game.__init__(self)
        self._players = [SinglePlayer(0, Character.default)]
        player = self._players[0]
        screen = pygame.display.get_surface()
        bkg = player.char.images["background"]
        try: pygame.mixer.music.fadeout(2000)
        except pygame.error: pass
        wipes.wipe_in(bkg)
        screen.blit(bkg, [0, 0])
        pygame.mixer.music.load(os.path.join("music", "single.ogg"))
        pygame.mixer.music.play(-1)
        player.dead = True
        self.play(0, [0, 0])
        self.block_out()

        text = "Game Over!"
        text = textfx.TextWavey(50, text, [255, 255, 255], amount = 5)

        statbox = player.win_data()
        statbox_r = statbox.get_rect(center = [player.centerx, 300])
        screen.blit(statbox, statbox_r)
        pygame.display.update()

        self.wait_reset(text, player.char)

        sec = "scores-single"
        score = player.score.score
        place = None
        if score > int(config.get(sec, "1").split(",")[1]):
            config.set(sec, "3", config.get(sec, "2"))
            config.set(sec, "2", config.get(sec, "1"))
            place = 1
        elif score > int(config.get(sec, "2").split(",")[1]):
            config.set(sec, "3", config.get(sec, "2"))
            place = 2
        elif score > int(config.get(sec, "3").split(",")[1]):
            place = 3

        if place:
            surf = player.char.border([500, 60])
            surf.blit(textfx.shadow("You got a high score! Enter your name:",
                                    26), [15, 10])
            string = u""
            while True:
                r = screen.blit(surf, [80, 440])
                screen.blit(textfx.shadow(string, 32), [100, 475])
                pygame.display.update(r)
                ev = pygame.event.wait()
                while ev.type != KEYDOWN: ev = pygame.event.wait()
                if ev.key == K_ESCAPE or ev.key == K_RETURN: break
                elif ev.key == K_BACKSPACE: string = string[:-1]
                elif ev.unicode == ",": pass
                elif ev.key < 128: string += ev.unicode

            if string == u"":
                string = os.environ.get("USER", "Ogi")
            config.set(sec, str(place), "%s,%d" % (string, score))
        check_unlock(player)

class VersusGame(Game):
    def __init__(self, chars):
        Game.__init__(self)
        matches = config.getint("settings", "matches")
        wins = [0] * len(chars)
        self._players = [VersusPlayer(i, c) for i,c in enumerate(chars)]
        screen = pygame.display.get_surface()

        for match in range(matches):
            bkg = random.choice([p.char.images["background"] for
                                 p in self._players])
            wipes.wipe_in(bkg)
            screen.blit(bkg, [0, 0])
            pygame.display.update()
            f = FightSprite(match, matches)
            f.add(self._sprites)
            try: pygame.mixer.music.fadeout(2000)
            except pygame.error: pass

            while self._sprites:
                self._sprites.update(pygame.time.get_ticks())
                pygame.display.update(self._sprites.draw(screen))
                self._sprites.clear(screen, bkg)
                pygame.time.clock.tick(60)

            pygame.mixer.music.load(os.path.join("music", "versus-1.ogg"))
            pygame.mixer.music.play(-1)
            self.play(matches, wins)
            if True not in [p.dead for p in self._players]: break
            else:
                self.block_out()

            try:
                winner = self._players[[p.dead
                                        for p in self._players].index(False)]
            except ValueError:
                winner = self._players[-1]
            winner.win()

            wins[winner._pid] += 1
            check_unlock(*self._players)
            if wins[winner._pid] > matches / 2:
                WinScreen(winner, matches)
                break

            text = "%s wins!" % (winner.char.name.split()[0])
            text = textfx.TextWavey(50, text, [255, 255, 255], amount = 5)

            statbox = winner.win_data()
            statbox_r = statbox.get_rect(center = [winner.centerx, 300])
            screen.blit(statbox, statbox_r)
            pygame.display.update()

            self.wait_reset(text, winner.char)

class ArcadeGame(Game):
    def __init__(self, chars):
        Game.__init__(self)
        matches = 3
        player = VersusPlayer(0, chars[0])
        screen = pygame.display.get_surface()

        ais = ["StupidAI", "VeryEasyAI", "EasyAI", "EasyAI",
               "NormalAI", "NormalAI", "HardAI", "InsaneAI"]

        quit = False
        for i, char in enumerate(Character.arcade):
            wins = [0] * len(chars)
            self._players = players = [player, AIPlayer(1, char, ais[i])]
            lost = False

            self._prebattle()

            for match in range(matches):
                bkg = random.choice([p.char.images["background"] for
                                     p in self._players])
                wipes.wipe_in(bkg)
                screen.blit(bkg, [0, 0])
                pygame.display.update()
                f = FightSprite(match, matches)
                f.add(self._sprites)
                try: pygame.mixer.music.fadeout(2000)
                except pygame.error: pass

                while self._sprites:
                    self._sprites.update(pygame.time.get_ticks())
                    pygame.display.update(self._sprites.draw(screen))
                    self._sprites.clear(screen, bkg)
                    pygame.time.clock.tick(60)

                pygame.mixer.music.load(os.path.join("music", "versus-1.ogg"))
                pygame.mixer.music.play(-1)
                self.play(matches, wins)
                if True not in [p.dead for p in self._players]:
                    quit = True
                    break
                else:
                    self.block_out()

                winner = list(filter(lambda p: not p.dead, self._players))[0]
                winner.win()
                check_unlock(self._players[0])
                wins[winner._pid] += 1
                if wins[winner._pid] > matches / 2:
                    if winner._pid == 0:
                        if i >= 3: config.unlock("cpuversus")
                        if i == 7: winner.score.score += 10000
                        WinScreen(winner, "arcade", to_score = (i == 7))
                    else:
                        WinScreen(winner, "arcade", scorer = self._players[0])
                        lost = True
                    break

                text = "%s wins!" % (winner.char.name.split()[0])
                text = textfx.TextWavey(50, text, [255, 255, 255], amount = 5)

                statbox = winner.win_data()
                statbox_r = statbox.get_rect(center = [winner.centerx, 300])
                screen.blit(statbox, statbox_r)
                pygame.display.update()

                self.wait_reset(text, winner.char)

            if lost or quit: break

    def _prebattle(self):
        screen = pygame.display.get_surface()
        em = EventManager()
        chars = [player.char for player in self._players]

        bkg = Character.default.images["background"].convert()
        bkg.blit(chars[0].border(chars[0].images["portrait"]), [20, 20])
        bkg.blit(chars[1].border(chars[1].images["portrait"]), [580, 390])

        txt1 = textfx.shadow(chars[0].name, 60)
        txt2 = textfx.shadow(chars[1].name, 60)
        bkg.blit(txt1, [230, 30])
        bkg.blit(txt2, txt2.get_rect(bottomright = [560, 580]))

        txt = textfx.shadow("versus", 40)
        bkg.blit(txt, txt.get_rect(center = [400, 300]))

        wipes.wipe_in(bkg)
        screen.blit(bkg, [0, 0])
        pygame.display.update()
        cont = True
        while cont:
            for ev in em.get():
                if (ev.type == QUIT or
                    (ev.type == PLAYER and
                     (ev.key in [CONFIRM, ROT_CC, ROT_CW]))):
                    cont = False
            pygame.time.clock.tick(60)

class AIGame(Game):
    def __init__(self, chars, all_ais = False, ai1 = None, ai2 = None):
        Game.__init__(self)
        matches = config.getint("settings", "matches")
        wins = [0] * len(chars)
        if all_ais:
            self._players = [AIPlayer(0, chars[0], ai1),
                             AIPlayer(1, chars[1], ai2)]
        else:
            self._players = [VersusPlayer(0, chars[0]),
                             AIPlayer(1, chars[1], ai1)]

        for match in range(matches):
            bkg = random.choice([p.char.images["background"] for
                                 p in self._players])
            screen = pygame.display.get_surface()
            wipes.wipe_in(bkg)
            screen.blit(bkg, [0, 0])
            pygame.display.update()
            f = FightSprite(match, matches)
            f.add(self._sprites)
            try: pygame.mixer.music.fadeout(2000)
            except pygame.error: pass

            while self._sprites:
                self._sprites.update(pygame.time.get_ticks())
                pygame.display.update(self._sprites.draw(screen))
                self._sprites.clear(screen, bkg)
                pygame.time.clock.tick(60)

            pygame.mixer.music.load(os.path.join("music", "versus-1.ogg"))
            pygame.mixer.music.play(-1)
            self.play(matches, wins)
            if True not in [p.dead for p in self._players]: break
            else:
                self.block_out()

            winner = self._players[[p.dead
                                    for p in self._players].index(False)]
            winner.win()

            wins[winner._pid] += 1
            if wins[winner._pid] > matches / 2:
                WinScreen(winner, matches, scorer = self._players[0])
                break

            text = "%s wins!" % (winner.char.name.split()[0])
            text = textfx.TextWavey(50, text, [255, 255, 255], amount = 5)

            statbox = winner.win_data()
            statbox_r = statbox.get_rect(center = [winner.centerx, 300])
            screen.blit(statbox, statbox_r)
            pygame.display.update()

            self.wait_reset(text, winner.char)

def WinScreen(winner, score_sec, to_score = True, scorer = None):
    em = EventManager()
    if scorer is None: scorer = winner
    screen = pygame.display.get_surface()
    print("Player %d wins!" % (winner._pid + 1))
    wipes.wipe_out()
    wipes.wipe_in(winner.char.images["background"])
    font = textfx.WrapFont(30, 400)
    text = font.render(random.choice(winner.char.quips))
    box = winner.char.border([text.get_width() + 20, text.get_height() + 20])
    statbox = winner.win_data()
    box.blit(text, [20, 20])
    screen.blit(winner.char.border(winner.char.images["portrait"]), [50, 50])
    screen.blit(box, [50, 350])
    screen.blit(statbox, [350, 50])
    pygame.display.update()

    if winner.is_ai:
        dead = load.sound("gameover.wav")
        dead.set_volume(0.5)
        dead.play()

    cont = True
    while cont:
        for ev in em.wait():
            if (ev.type == QUIT or
                (ev.type == PLAYER and
                 (ev.key in [CONFIRM, ROT_CC, ROT_CW]))):
                cont = False

    score = scorer.score.score
    place = None

    if not scorer.is_ai and to_score:
        sec = "scores-versus-%s" % str(score_sec)
        if score > int(config.get(sec, "1").split(",")[1]):
            config.set(sec, "3", config.get(sec, "2"))
            config.set(sec, "2", config.get(sec, "1"))
            place = 1
        elif score > int(config.get(sec, "2").split(",")[1]):
            config.set(sec, "3", config.get(sec, "2"))
            place = 2
        elif score > int(config.get(sec, "3").split(",")[1]):
            place = 3

    if not scorer.is_ai and to_score and place:
        surf = scorer.char.border([500, 60])
        surf.blit(textfx.shadow("You got a high score! Enter your name:", 26),
                  [15, 10])
        string = u""
        while True:
            r = screen.blit(surf, [80, 440])
            screen.blit(textfx.shadow(string + "_", 32), [100, 475])
            pygame.display.update(r)
            ev = pygame.event.wait()
            while ev.type != KEYDOWN: ev = pygame.event.wait()
            if ev.key == K_ESCAPE or ev.key == K_RETURN: break
            elif ev.key == K_BACKSPACE: string = string[:-1]
            elif ev.unicode == ",": pass
            elif ev.key < 128: string += ev.unicode

        if string == u"":
            string = os.environ.get("USER", scorer.char.name)
        config.set(sec, str(place), "%s,%d" % (string, score))

def check_unlock(*players):
    for p in filter(lambda p: not p.is_ai, players):
        if p.field.max_gemsize >= 25:
            config.unlock("unixbros")
        if p.field.max_chain >= 4:
            config.unlock("combat")



#!/usr/local/bin/python3.9

# angrydd - a falling block puzzle game
# Copyright 2004 Joe Wreschnig <piman@sacredchao.net>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# version 2 along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

__date__ = "$LastChangedDate: 2006-01-12 15:19:27 -0600 (Thu, 12 Jan 2006) $"
__version__   = "1.0.1"
__revision__  = "$Id: angrydd.py 317 2006-01-12 21:19:27Z piman $"
__copyright__ = "Copyright 2004 Joe Wreschnig <piman@sacredchao.net>"
__license__   = "GNU GPL v2"
__author__    = "Joe Wreschnig <piman@sacredchao.net>"

import os
import sys

def print_help():
    print("""\
Usage: %s [ --debug | --help | --version ]
Options:
  --help, -h     Print this help message and exit.
  --version, -v  Print version and copyright information and exit.
  --debug        Start without Psyco support, for profiling and debugging.
  --profile      Start with profiling (implies --debug).
""" %(
        sys.argv[0]))
    raise SystemExit

def print_version():
    print("""\
Angry Drunken Dwarves %s - a falling block puzzle game
Copyright 2004 %s

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU GPL version 2, as published
by the Free Software Foundation.""" % (__version__, __author__))
    raise SystemExit

if __name__ == "__main__":
    if ("--help" in sys.argv) or ("-h" in sys.argv): print_help()
    elif ("--version" in sys.argv) or ("-v" in sys.argv): print_version()

angrydd_path = os.path.split(os.path.realpath(__file__))[0]
if os.path.exists(os.path.join(angrydd_path, "angrydd.zip")):
    sys.path.insert(0, os.path.join(angrydd_path, "angrydd.zip"))

import pygame
from pygame import mixer

from constants import *

import config
import load

# Toggling fullscreen doesn't work on non-X11 platforms as of Pygame 1.7.
def toggle_fs():
    flags = 0
    fs = config.getboolean("settings", "fullscreen") ^ True
    if "linux" not in sys.platform:
        if fs: flags ^= FULLSCREEN
        pygame.display.set_mode([800, 600], flags)
    else: pygame.display._toggle_fullscreen()
    config.set("settings", "fullscreen", str(fs))

pygame.display._toggle_fullscreen = pygame.display.toggle_fullscreen
pygame.display.toggle_fullscreen = toggle_fs

def main():
    from characters import Character

    init()

    import menu
    import howtoplay
    import unlocker
    main_menu = [
        ("Dwarf versus Dwarf", menu.entry(play_game, False)),
        ("Arcade Mode", menu.entry(play_arcade_game, False)),
        ("Setup", menu.entry(setup, False)),
        ("How to Play", menu.entry(howtoplay.init, False)),
        ("Quit Game", menu.break_menu)]

    if config.getboolean("unlock", "unlocker"):
        main_menu.insert(3, ("View Unlocks", menu.entry(unlocker.init, False)))

    if config.getboolean("unlock", "single"):
        main_menu.insert(2, ("Single Player",
                             menu.entry(play_single_game, False)))
    if config.getboolean("unlock", "cpuversus"):
        main_menu.insert(2,
                         ("Dwarf versus Robot",
                          menu.entry(play_ai_game, False)))

    if len(Character.arcade) != 8:
        main_menu = [main_menu[0]] + main_menu[2:]

    menu.Menu(main_menu)
    quit()

# Start a player vs. player game
def play_game(menu, platform, pos, key):
    import game, charselect
    characters = charselect.init()
    if characters[0] is not None:
        game.VersusGame(characters)
        mixer.music.load(os.path.join("music", "intro.ogg"))
        mixer.music.play(-1)
    return True

# Start a single player game
def play_single_game(menu, platform, pos, key):
    import game
    game.SingleGame()
    mixer.music.load(os.path.join("music", "intro.ogg"))
    mixer.music.play(-1)
    return True

# Start a game versus an AI
def play_ai_game(menu, platform, pos, key):
    import game, charselect
    characters = charselect.init()
    if characters[0] is not None:
        game.AIGame(characters)
        mixer.music.load(os.path.join("music", "intro.ogg"))
        mixer.music.play(-1)
    return True

def play_arcade_game(menu, platform, pos, key):
    import game, charselect
    characters = charselect.init(numplayers = 1)
    if characters[0] is not None:
        game.ArcadeGame(characters)
        mixer.music.load(os.path.join("music", "intro.ogg"))
        mixer.music.play(-1)
    return True

# Open the configuration menu
def setup(menu_, platform, pos, key):
    import menu
    config_menu = [
        (config.get_matches(), menu.entry(config.set_matches)),
        (config.get_speed(), menu.entry(config.set_speed)),
        (config.get_rotup(), menu.entry(config.set_rotup)),
        (config.get_space(), menu.entry(config.set_space)),

        ("Back", menu.break_menu)
        ]

    if config.getboolean("unlock", "cpuversus"):
        config_menu.insert(2, (config.get_ai(), menu.entry(config.set_ai)))

    if config.getboolean("unlock", "combat"):
        config_menu.insert(-1, (config.get_combat(),
                                menu.entry(config.set_combat)))

    try: menu.Menu(config_menu)
    except menu.MenuExit: pass
    return True

# Set up Pygame, load sounds, etc.
def init():
    if os.name == 'posix': # We need to force stereo in many cases.
        try: mixer.pre_init(44100, -16, 2)
        except pygame.error: pass

    pygame.init()
    config.init(rc_file)
    os.chdir(angrydd_path)
    pygame.display.set_caption("Angry, Drunken Dwarves")
    try: pygame.display.set_icon(pygame.image.load("angrydd.png"))
    except pygame.error: pass
    pygame.mouse.set_visible(False)
    if config.getboolean("settings", "fullscreen"): flags = FULLSCREEN
    else: flags = 0
    pygame.display.set_mode([800, 600], flags)

    import game
    import menu
    import boxes
    import characters

    boxes.TickBox.sound = load.sound("tick.wav")
    boxes.TickBox.sound.set_volume(0.3)

    boxes.BreakBox.sound = load.sound("break.wav")
    boxes.BreakBox.sound.set_volume(0.3)

    boxes.Special.sound = load.sound("select-confirm.wav")
    boxes.Special.sound.set_volume(0.5)

    game.FallingBoxes.rotate_sound = load.sound("rotate.wav")
    game.FallingBoxes.rotate_sound.set_volume(0.2)

    menu.Menu.bg = load.image("menu-bg.png").convert()

    characters.init()

    mixer.music.load(os.path.join("music", "intro.ogg"))
    mixer.music.play(-1)

# Shut down Pygame, save configuration data, turn on the Unlocker.
def quit(*args):
    pygame.quit()
    print("Saving configuration.")
    config.unlock("unlocker")
    if int(config.get("scores-versus-3", "1").split(",")[1]) > 20000:
        config.unlock("single")
    config.write(open(rc_file, "w"))
    raise SystemExit

if __name__ == "__main__":
    profile = ("--profile" in sys.argv)
    debug = profile or ("--debug" in sys.argv)
    double_ai = ("--double-ai" in sys.argv)

    if not debug:
        try:
            import psyco
            psyco.full()
            print("Psyco optimizing compiler found. Using psyco.full().")
        except ImportError: pass

    if profile:
        import profile
        profile.run("main()")
    elif double_ai:
        # Hack to run two AIs versus each other; use --double-ai AI1,Ai2
        # on the command line. Bypasses main(), very buggy, for testing
        # use only.
        init()
        ai1, ai2 = sys.argv[sys.argv.index("--double-ai") + 1].split(",")
        game.AIGame([Character.available[0], Character.available[0]],
                     all_ais = True, ai1 = ai1, ai2 = ai2)
    else: main()



# wipes.py - a set of simple screen wipes (in and out)
# Copyright 2004 Joe Wreschnig <piman@sacredchao.net>
# Released under the terms of the GNU GPL v2.
__revision__ = "$Id: wipes.py 286 2004-09-04 03:51:59Z piman $"

import random
from pygame import display, draw, time, Rect, Surface
from itertools import chain

c = time.Clock()

def line_out_l2r():
    screen = display.get_surface()
    w, h = screen.get_size()
    for x in chain(range(0, w, 2), range(w - 1, 0, -2)):
        display.update(draw.line(screen, [0, 0, 0], [x, 0], [x, h]))
        c.tick(700)

def line_out_r2l():
    screen = display.get_surface()
    w, h = screen.get_size()
    for x in chain(range(w - 1, 0, -2), range(0, w, 2)):
        display.update(draw.line(screen, [0, 0, 0], [x, 0], [x, h]))
        c.tick(700)

def line_out_t2b():
    screen = display.get_surface()
    w, h = screen.get_size()
    for y in chain(range(0, h, 2), range(h - 1, 0, -2)):
        display.update(draw.line(screen, [0, 0, 0], [0, y], [w, y]))
        c.tick(700)

def line_out_b2t():
    screen = display.get_surface()
    w, h = screen.get_size()
    for y in chain(range(h - 1, 0, -2), range(0, h, 2)):
        display.update(draw.line(screen, [0, 0, 0], [0, y], [w, y]))
        c.tick(700)

def line_in_l2r(surf):
    screen = display.get_surface()
    w, h = screen.get_size()
    for x in chain(range(0, w, 2), range(w - 1, 0, -2)):
        display.update(screen.blit(surf, [x, 0, 1, h], [x, 0, 1, h]))
        c.tick(700)

def line_in_r2l(surf):
    screen = display.get_surface()
    w, h = screen.get_size()
    for x in chain(range(w - 1, 0, -2), range(0, w, 2)):
        display.update(screen.blit(surf, [x, 0, 1, h], [x, 0, 1, h]))
        c.tick(700)

def line_in_t2b(surf):
    screen = display.get_surface()
    w, h = screen.get_size()
    for y in chain(range(0, h, 2), range(h - 1, 0, -2)):
        display.update(screen.blit(surf, [0, y, w, 1], [0, y, w, 1]))
        c.tick(700)

def line_in_b2t(surf):
    screen = display.get_surface()
    w, h = screen.get_size()
    for y in chain(range(h - 1, 0, -2), range(0, h, 2)):
        display.update(screen.blit(surf, [0, y, w, 1], [0, y, w, 1]))
        c.tick(700)

OUTS = [line_out_l2r, line_out_r2l, line_out_b2t, line_out_t2b]
INS = [line_in_l2r, line_in_r2l, line_in_b2t, line_in_t2b]

def wipe_out(): random.choice(OUTS)()
def wipe_in(s): random.choice(INS)(s)
